from datetime import datetime
from typing import List, Optional

from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class HeiseParser(ParserProxy):
    class V1(BaseParser):
        _summary_selector = XPath(
            "//article[not(@data-component='TeaserContainer')]//p[@class='a-article-header__lead'] | //article[not(@data-component='TeaserContainer')]//div[@id='lead']/p"
        )
        _subheadline_selector = XPath(
            "//article[not(@data-component='TeaserContainer')]//h3[contains(@class,'subheading')] | //article[not(@data-component='TeaserContainer')]//h2[@class='heading-h2 replaced-h1']"
        )
        _paragraph_selector = XPath(
            "//div[contains(class, article-content)]/p[not(@class"
            " or ((string-length(text()) < 3) and (contains(text(), '(') or contains(span, '(')))"
            " or contains(text(), '=== Anzeige / Sponsorenhinweis')"
            " or contains(text(), 'Tipp: Wir sind bei WhatsApp!')"
            " or contains(a, 'heise+ abonnieren')"
            " or contains(text(), '► '))"
            " or @class='antwort rte__abs--antwort'"
            " or @class='frage rte__abs--frage'"
            " or @class='json-ld-paid-content-marker'] "
            " | //div[contains(class, article-content)]//ul["
            "@class='rte__list rte__list--unordered' or @class='boxtext']/li | "
            # The selectors below this line are specific for heise's blog: techstage.de
            "//div[@class='ringCommonDetail ringBlockType-paragraph ']/p"
            " | //div[@class='ringCommonDetail ringBlockType-paragraph ']//ul/li"
        )

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def publishing_date(self) -> Optional[datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("keywords"))

        @attribute
        def images(self) -> List[Image]:
            # There some (rare) cases there are some images that are not being extracted with this, because they are
            # referenced by relative URLs. e.g.
            # https://www.heise.de/hintergrund/Zahlen-bitte-136199-Eris-Der-Grund-warum-Pluto-kein-Planet-mehr-ist-9993800.html
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath(
                    "//h1[@class='article-headline ' or contains(@class, 'a-article-header__title')]"
                ),
                image_selector=XPath(
                    "//div[@class='article-image__gallery-container']//img|"
                    "//div[@class='image-container']//img|"
                    "//div[@class='article-layout__content']//figure[not(@class)]//noscript//img"
                ),
                caption_selector=XPath(
                    "./ancestor::figure//p[@class='a-caption__text']|"
                    "./ancestor::figure//div[@class='text']|"
                    "./ancestor::div[@class='article-gallery ']//span[@class='caption']"
                ),
                author_selector=XPath(
                    "./ancestor::figure//p[@class='a-caption__source']|"
                    "./ancestor::div[@class='article-gallery ']//span[@class='copyright']"
                ),
            )



import datetime
import re
from typing import List, Optional

from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    image_extraction,
)


class WochenblattParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = XPath("//div[@class='entry-content']/p[position() < last() and not(b)]")
        _subheadline_selector = XPath(
            "//div[@class='entry-content']/*[(self::p or self::h3) and position() < last() and b]"
        )

        _author_selector = XPath("//div[@class='entry-content']/p[last()]")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute(priority=1)
        def authors(self) -> List[str]:
            # The author is only listed in the last line of the article
            authors = self._author_selector(self.precomputed.doc)
            if authors:
                author = authors[0].text_content().strip()
                if match := re.match(r"(?i)^wochenblatt\s*/\s*(?P<authors>([\w ]+))", author):
                    return generic_author_parsing(match.group("authors"))
            return []

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return re.sub(r"(?i)\s*-\s*wochenblatt", "", self.precomputed.meta.get("og:title") or "")

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//div[@class='entry-content']//img"),
            )



import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
)


class RBB24Parser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = XPath("//div[contains(concat(' ', @class , ' '), ' textblock ')]/p")
        _summary_selector = XPath("//div[contains(concat(' ', @class , ' '), ' shorttext ')]/p")
        _subheadline_selector = XPath("//h4[contains(concat(' ', @class , ' '), ' texttitle ')]")
        _author_selector = CSSSelector("span.authorname")
        _date_selector = CSSSelector("div.lineinfo")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            article_body = extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                paragraph_selector=self._paragraph_selector,
                subheadline_selector=self._subheadline_selector,
            )
            # If author was specified, the author is often credited after the last sentence of the summary:
            # "... . Von Max Mustermann". We need to remove this.
            # As XPath can only select whole subtrees, we need to remove it later,
            # because the html will look like:
            # <div class="shorttext">
            #   <p> Summary Text
            #     <i>Von Max Mustermann</i>
            #   </p>
            # </div>
            #
            # Check if author was specified and if the author was also credited:
            if len(self._author_selector(self.precomputed.doc)) > 0:
                # Get authors
                authors_list = self.authors()
                # Get summary as string
                summary = article_body.summary._data[0]
                # Make sure that we choose a period that is not the last character of string,
                # in case it was written like: "... . Von Max Mustermann."
                potential_end_of_summary = summary[: len(summary) - 1].rfind(".")
                potential_credits = summary[potential_end_of_summary + 2 : len(summary)]
                # Check whether the last sentence credits the author
                # Does the sentence start with "Von"
                if potential_credits.split(" ", 1)[0] == "Von":
                    authors_credited = True
                    # Are the authors credited ?
                    for author in authors_list:
                        if author not in potential_credits:
                            authors_credited = False
                            break
                    # Delete last sentence if this sentence just credits the authors
                    if authors_credited:
                        # As the string of the summary is a tuple we can't change it and need to create a new one
                        article_body.summary._data = (summary[: potential_end_of_summary + 1],)
            # Often articles end with "Sendung: rbb24 Abendschau, 30.04.2024, 19:30 Uhr" or
            # "Sendung: Der Tag, 26.04.2024, 19:15 Uhr" etc
            # We need to delete this last paragraph
            last_paragraph = article_body.sections[-1].paragraphs._data[-1]
            if last_paragraph.startswith("Sendung:"):
                # Make tuple to list to delete last element of list, to then make it a tuple again
                new_data_tuple = (article_body.sections[-1].paragraphs._data)[:-1]
                article_body.sections[-1].paragraphs._data = new_data_tuple
            return article_body

        @attribute
        def authors(self) -> List[str]:
            # check if author was specified
            if len(self._author_selector(self.precomputed.doc)) > 0:
                authors = self._author_selector(self.precomputed.doc)[0].text
                return generic_author_parsing(authors)
            else:
                return []

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            publishing_date_string = self._date_selector(self.precomputed.doc)[0].text
            # publishing_date_string will look like 'Do 25.04.24 | 13:47 Uhr'
            # need to get the date and time to pass it to the generic_date_parsing
            if publishing_date_string is not None:
                index = publishing_date_string.index("|")
                date_string = publishing_date_string[index - 9 : index - 1]
                time_string = publishing_date_string[index + 1 : index + 7]
                return generic_date_parsing(date_string + " " + time_string)
            else:
                return generic_date_parsing(None)

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, ParserProxy, attribute
from fundus.parser.base_parser import function
from fundus.parser.data import Image
from fundus.parser.utility import (
    extract_article_body_with_selector,
    extract_json_from_dom,
    generic_author_parsing,
    generic_date_parsing,
    image_extraction,
)


class CBCNewsParser(ParserProxy):
    class V1(BaseParser):
        _summary_selector = CSSSelector("h2.deck")
        _subheadline_selector = CSSSelector("div.story > h2")
        _paragraph_selector = CSSSelector("div.story > p")

        _cbc_ld_selector: XPath = XPath("//script[@id='initialStateDom']")

        @function(priority=1)
        def _parse_initial_state_dom(self):
            state_dom_json = extract_json_from_dom(self.precomputed.doc, self._cbc_ld_selector)
            for ld in state_dom_json:
                self.precomputed.ld.add_ld(ld, "initialStateDom")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def topics(self) -> List[str]:
            if not (topic_dict := self.precomputed.ld.bf_search("keywords")):
                return []

            # add locations
            topic_list = [topic for location in topic_dict.get("tags") if (topic := location.get("name")) is not None]

            # add subjects
            for subject in topic_dict.get("concepts"):
                if (path := subject.get("path")) is not None:
                    topic_list.append(re.sub(r".*/", "", path))

            return topic_list

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath("//div[@data-cy='storyWrapper']"),
                caption_selector=XPath(
                    "./ancestor::figure//figcaption | ./ancestor::span[contains(@class,'mediaEmbed')]/span"
                ),
                author_selector=re.compile(r"\((?P<credits>.*?)\)$"),
            )



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class MorgunbladidParser(ParserProxy):
    class V1(BaseParser):
        _summary_selector = XPath("//div[@class='main-layout']//div[@class='is-merking']/p")
        _paragraph_selector = XPath(
            "//div[@class='main-layout' or @data-element-type='body-facts']" "/p[not(a and not(text()))]"
        )
        _subheadline_selector = XPath("//div[@class='main-layout' or @class='et_pb_text_inner']/h3")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//div[@class='image']//img"),
                caption_selector=XPath("./ancestor::div[contains(@class, 'newsitem-image')]//span[@class='caption']"),
                author_selector=XPath("./ancestor::div[contains(@class, 'newsitem-image')]//span[@class='credit']"),
            )



import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    image_extraction,
    normalize_whitespace,
)


class TheBBCParser(ParserProxy):
    class V1(BaseParser):
        _subheadline_selector = XPath(
            "//div[@data-component='subheadline-block' or @data-component='text-block' or contains(@class, 'ebmt73l0')]//*[self::h2 or (self::p and b and not(text()) and position()>1)]"
        )
        _summary_selector = XPath(
            "(//div[@data-component='text-block' or contains(@class, 'ebmt73l0')])[1] //p[b and not(text) and position()=1]"
        )
        _paragraph_selector = XPath(
            "//div[@data-component='text-block' or contains(@class, 'ebmt73l0')][1]//p[not(b) and text()] |"
            "//div[@data-component='text-block' or contains(@class, 'ebmt73l0')][position()>1] //p[text()] |"
            "//div[@data-component='text-block' or contains(@class, 'ebmt73l0')] //ul /li[text()]"
        )

        _topic_selector = CSSSelector(
            "div[data-component='topic-list'] > div > div > ul > li ," "div[data-component='tags'] a"
        )

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                subheadline_selector=self._subheadline_selector,
                summary_selector=self._summary_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def topics(self) -> List[str]:
            topic_nodes = self._topic_selector(self.precomputed.doc)
            return [normalize_whitespace(node.text_content()) for node in topic_nodes]

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//figure //img[not(@src='/bbcx/grey-placeholder.png')]"),
                caption_selector=XPath("./ancestor::figure//figcaption//p"),
                author_selector=XPath("./ancestor::figure//span[@role='text']/text()"),
            )



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class TechCrunchParser(ParserProxy):
    class V1(BaseParser):
        VALID_UNTIL = datetime.date(2024, 5, 8)
        _summary_selector: XPath = CSSSelector("div.article-content > p#speakable-summary")
        _paragraph_selector: XPath = CSSSelector(
            "div.article-content > p:not(#speakable-summary), div.article-content > ul > li"
        )
        _subheadline_selector: CSSSelector = CSSSelector("div.article-content > h2")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            body: ArticleBody = extract_article_body_with_selector(
                self.precomputed.doc,
                subheadline_selector=self._subheadline_selector,
                summary_selector=self._summary_selector,
                paragraph_selector=self._paragraph_selector,
            )
            return body

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.meta.get("sailthru.author"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            if headline := self.precomputed.meta.get("og:title"):
                return headline.replace(" | TechCrunch", "")
            return None

        @attribute
        def topics(self) -> List[str]:
            if topics := generic_topic_parsing(self.precomputed.meta.get("keywords")):
                return topics
            else:
                return generic_topic_parsing(self.precomputed.meta.get("sailthru.tags"))

    class V1_1(V1):
        VALID_UNTIL = datetime.date.today()
        _summary_selector: XPath = CSSSelector("div.entry-content > p#speakable-summary")
        _paragraph_selector: XPath = CSSSelector(
            "div.entry-content > p:not(#speakable-summary), div.entry-content > ul > li"
        )
        _subheadline_selector: CSSSelector = CSSSelector("div.entry-content > h2")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            body: ArticleBody = extract_article_body_with_selector(
                self.precomputed.doc,
                subheadline_selector=self._subheadline_selector,
                summary_selector=self._summary_selector,
                paragraph_selector=self._paragraph_selector,
            )
            return body

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.meta.get("author"))

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.ld.bf_search("keywords"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath(
                    "//div[@class='article-hero__first-section']|"
                    "//div[@class='is-floating wp-block-techcrunch-social-share']|"
                    "//h1[@class='wp-block-post-title']"
                ),
                caption_selector=XPath("./ancestor::figure//figcaption"),
                author_selector=re.compile(r"(?i)image credits:(?P<credits>.*)"),
            )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.publishers.ind.bhaskar import BhaskarParser
from fundus.publishers.ind.times_of_india import TimesOfIndiaParser
from fundus.scraping.url import NewsMap, RSSFeed


class IND(metaclass=PublisherGroup):
    default_language = "hi"

    TimesOfIndia = Publisher(
        name="Times Of India",
        domain="https://www.timesofindia.indiatimes.com",
        parser=TimesOfIndiaParser,
        sources=[
            NewsMap("https://timesofindia.indiatimes.com/sitemap/today", languages={"en"}),
            NewsMap("https://timesofindia.indiatimes.com/sitemap/yesterday", languages={"en"}),
            RSSFeed("https://timesofindia.indiatimes.com/rssfeedstopstories.cms", languages={"en"}),
            RSSFeed("https://timesofindia.indiatimes.com/rssfeedmostrecent.cms", languages={"en"}),
        ],
    )

    Bhaskar = Publisher(
        name="Dainik Bhaskar",
        domain="https://www.bhaskar.com/",
        parser=BhaskarParser,
        sources=[NewsMap("https://www.bhaskar.com/sitemaps-v1--sitemap-google-news-index.xml")],
    )



import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
    parse_title_from_root,
    strip_nodes_to_text,
)


class NetzpolitikOrgParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = CSSSelector("div.entry-content p")
        _summary_selector = CSSSelector("div.entry-excerpt > p")
        _subheadline_selector = CSSSelector("div.entry-content > h3")
        _author_selector = CSSSelector("span > a[rel='author']")
        _topic_selector = CSSSelector("div.entry-footer__tags li")

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title") or parse_title_from_root(self.precomputed.doc)

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def topics(self) -> List[str]:
            topic_string = strip_nodes_to_text(self._topic_selector(self.precomputed.doc), join_on=",")
            if topic_string is not None:
                return generic_topic_parsing(topic_string, delimiter=",")
            return []

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.meta.get("article:published_time"))

        @attribute
        def authors(self) -> List[str]:
            return [node.text_content() for node in (self._author_selector(self.precomputed.doc) or [])]

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                caption_selector=XPath("./ancestor::figure//figcaption/text()"),
                author_selector=XPath("./ancestor::figure//figcaption/span"),
            )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.filter import inverse, lor, regex_filter
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap

from .ap_news import APNewsParser
from .business_insider import BusinessInsiderParser
from .cnbc import CNBCParser
from .fox_news import FoxNewsParser
from .free_beacon import FreeBeaconParser
from .la_times import LATimesParser
from .occupy_democrats import OccupyDemocratsParser
from .reuters import ReutersParser
from .rolling_stone import RollingStoneParser
from .techcrunch import TechCrunchParser
from .the_gateway_pundit import TheGatewayPunditParser
from .the_intercept import TheInterceptParser
from .the_nation import TheNationParser
from .the_new_yorker import TheNewYorkerParser
from .voice_of_america import VOAParser
from .washington_post import WashingtonPostParser
from .washington_times import WashingtonTimesParser
from .wired import WiredParser
from .world_truth import WorldTruthParser


class US(metaclass=PublisherGroup):
    default_language = "en"

    APNews = Publisher(
        name="Associated Press News",
        domain="https://apnews.com/",
        parser=APNewsParser,
        sources=[
            Sitemap(
                "https://apnews.com/sitemap.xml",
                sitemap_filter=regex_filter("apnews.com/hub/|apnews.com/video/"),
                reverse=True,
                languages={"en", "es"},
            ),
            NewsMap("https://apnews.com/news-sitemap-content.xml", languages={"en", "es"}),
        ],
    )

    CNBC = Publisher(
        name="CNBC",
        domain="https://www.cnbc.com/",
        parser=CNBCParser,
        sources=[
            Sitemap("https://www.cnbc.com/sitemapAll.xml"),
            NewsMap("https://www.cnbc.com/sitemap_news.xml"),
        ],
    )

    TechCrunch = Publisher(
        name="TechCrunch",
        domain="https://techcrunch.com/",
        parser=TechCrunchParser,
        sources=[
            Sitemap(
                "https://techcrunch.com/sitemap_index.xml",
                sitemap_filter=inverse(regex_filter("post-sitemap")),
                reverse=True,
            ),
            NewsMap("https://techcrunch.com/news-sitemap.xml"),
        ],
    )

    TheIntercept = Publisher(
        name="The Intercept",
        domain="https://theintercept.com/",
        parser=TheInterceptParser,
        sources=[
            RSSFeed("https://theintercept.com/feed/?lang=en"),
            Sitemap(
                "https://theintercept.com/sitemap_index.xml",
                reverse=True,
                sitemap_filter=inverse(regex_filter("post-sitemap")),
            ),
        ],
    )

    TheGatewayPundit = Publisher(
        name="The Gateway Pundit",
        domain="https://www.thegatewaypundit.com/",
        parser=TheGatewayPunditParser,
        sources=[
            Sitemap(
                "https://www.thegatewaypundit.com/sitemap_index.xml",
                sitemap_filter=inverse(regex_filter("post-sitemap")),
                reverse=True,
            ),
            NewsMap("https://www.thegatewaypundit.com/news-sitemap.xml"),
        ],
    )

    FoxNews = Publisher(
        name="Fox News",
        domain="https://www.foxnews.com/",
        parser=FoxNewsParser,
        url_filter=regex_filter(r"\/video\/"),
        sources=[
            Sitemap(
                "https://www.foxnews.com/sitemap.xml",
                sitemap_filter=inverse(regex_filter("type=articles")),
            ),
            NewsMap("https://www.foxnews.com/sitemap.xml?type=news"),
            RSSFeed("https://moxie.foxnews.com/google-publisher/latest.xml"),
            RSSFeed("https://moxie.foxnews.com/google-publisher/world.xml"),
            RSSFeed("https://moxie.foxnews.com/google-publisher/politics.xml"),
            RSSFeed("https://moxie.foxnews.com/google-publisher/us.xml"),
            RSSFeed("https://moxie.foxnews.com/google-publisher/world.xml"),
            RSSFeed("https://moxie.foxnews.com/google-publisher/travel.xml"),
            RSSFeed("https://moxie.foxnews.com/google-publisher/opinion.xml"),
            RSSFeed("https://moxie.foxnews.com/google-publisher/tech.xml"),
            RSSFeed("https://moxie.foxnews.com/google-publisher/science.xml"),
            RSSFeed("https://moxie.foxnews.com/google-publisher/health.xml"),
        ],
    )

    TheNation = Publisher(
        name="The Nation",
        domain="https://www.thenation.com/",
        parser=TheNationParser,
        sources=[
            Sitemap(
                "https://www.thenation.com/sitemap_index.xml",
                sitemap_filter=inverse(regex_filter("article-sitemap")),
                reverse=True,
            ),
            NewsMap("https://www.thenation.com/news-sitemap.xml"),
        ],
    )

    # WorldTruth = Publisher(
    #     name="World Truth",
    #     domain="https://www.worldtruth.tv/",
    #     sources=[RSSFeed("https://feeds.feedburner.com/ConsciousnessTv")],
    #     parser=WorldTruthParser,
    # )

    FreeBeacon = Publisher(
        name="The Washington Free Beacon",
        domain="https://freebeacon.com/",
        sources=[
            Sitemap(
                "https://freebeacon.com/wp-sitemap.xml",
                sitemap_filter=inverse(regex_filter("post-sitemap")),
                reverse=True,
            ),
            Sitemap(
                "https://freebeacon.com/wp-sitemap.xml",
                sitemap_filter=inverse(regex_filter("blog-sitemap")),
                reverse=True,
            ),
        ],
        parser=FreeBeaconParser,
    )

    WashingtonTimes = Publisher(
        name="The Washington Times",
        domain="https://www.washingtontimes.com/",
        parser=WashingtonTimesParser,
        sources=[
            RSSFeed("https://www.washingtontimes.com/rss/headlines/news/politics/"),
            Sitemap("https://www.washingtontimes.com/sitemap-stories.xml"),
            Sitemap("https://www.washingtontimes.com/sitemap-entries.xml"),
        ],
        deprecated=True,
    )

    WashingtonPost = Publisher(
        name="Washington Post",
        domain="https://www.washingtonpost.com/",
        parser=WashingtonPostParser,
        sources=[
            Sitemap("https://www.washingtonpost.com/sitemaps/sitemap.xml.gz"),
            NewsMap("https://www.washingtonpost.com/sitemaps/news-sitemap.xml.gz"),
            RSSFeed("https://feeds.washingtonpost.com/rss/world"),
            RSSFeed("https://feeds.washingtonpost.com/rss/national"),
        ],
        # Adds a URL-filter to ignore incomplete URLs
        url_filter=regex_filter(r"washingtonpost.com(\/)?$"),
        deprecated=True,
    )

    TheNewYorker = Publisher(
        name="The New Yorker",
        domain="https://www.newyorker.com/",
        parser=TheNewYorkerParser,
        sources=[
            Sitemap("https://www.newyorker.com/sitemap.xml"),
            NewsMap("https://www.newyorker.com/feed/google-news-sitemap-feed/sitemap-google-news"),
        ],
    )

    Reuters = Publisher(
        name="Reuters",
        domain="https://www.reuters.com/",
        parser=ReutersParser,
        sources=[
            Sitemap("https://www.reuters.com/arc/outboundfeeds/sitemap-index/?outputType=xml"),
            NewsMap("https://www.reuters.com/arc/outboundfeeds/news-sitemap-index/?outputType=xml"),
        ],
        deprecated=True,
    )

    OccupyDemocrats = Publisher(
        name="Occupy Democrats",
        domain="https://occupydemocrats.com/",
        parser=OccupyDemocratsParser,
        sources=[
            Sitemap(
                url="https://occupydemocrats.com/sitemap.xml",
                sitemap_filter=inverse(regex_filter(r"post-sitemap")),
            )
        ],
        deprecated=True,
    )

    LATimes = Publisher(
        name="Los Angeles Times",
        domain="https://www.latimes.com/",
        parser=LATimesParser,
        sources=[
            Sitemap("https://www.latimes.com/sitemap.xml"),
            NewsMap("https://www.latimes.com/news-sitemap.xml"),
        ],
    )

    BusinessInsider = Publisher(
        name="Business Insider",
        domain="https://www.businessinsider.com/",
        parser=BusinessInsiderParser,
        sources=[
            NewsMap("https://www.businessinsider.com/sitemap/google-news.xml"),
            Sitemap("https://www.businessinsider.com/sitemap/2024-01.xml"),
        ],
    )

    RollingStone = Publisher(
        name="Rolling Stone",
        domain="https://www.rollingstone.com/",
        parser=RollingStoneParser,
        sources=[
            NewsMap("https://www.rollingstone.com/news-sitemap.xml"),
            Sitemap(
                "https://www.rollingstone.com/sitemap_index.xml",
                sitemap_filter=inverse(lor(regex_filter("/pmc_list-sitemap"), regex_filter("/post-sitemap"))),
            ),
        ],
    )

    VoiceOfAmerica = Publisher(
        name="Voice Of America",
        domain="https://www.voanews.com/",
        parser=VOAParser,
        url_filter=inverse(regex_filter(r"voanews\.com\/a\/[a-z-]+\/[0-9]+\.html")),
        sources=[
            Sitemap(
                "https://www.voanews.com/sitemap.xml",
                sitemap_filter=inverse(regex_filter(r"sitemap_[\d_]*\.xml\.gz")),
            ),
        ],
    )

    Wired = Publisher(
        name="Wired",
        domain="https://www.wired.com",
        parser=WiredParser,
        sources=[
            RSSFeed("https://www.wired.com/feed/rss"),
            NewsMap("https://www.wired.com/feed/google-latest-news/sitemap-google-news"),
            Sitemap("https://www.wired.com/sitemap.xml"),
            Sitemap("https://www.wired.com/sitemap-archive-1.xml"),
        ],
    )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import Sitemap

from .mexico_news_daily import MexicoNewsDailyParser


class MX(metaclass=PublisherGroup):
    default_language = "es"

    MexicoNewsDaily = Publisher(
        name="Mexico News Daily",
        domain="https://mexiconewsdaily.com/",
        parser=MexicoNewsDailyParser,
        sources=[
            Sitemap(
                "https://mexiconewsdaily.com/sitemap_index.xml",
                sitemap_filter=inverse(regex_filter(r"post-sitemap")),
                languages={"en"},
            )
        ],
    )



import datetime
from typing import Optional

from lxml.cssselect import CSSSelector

from fundus.parser import ArticleBody, BaseParser, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_date_parsing,
)


class WorldTruthParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = CSSSelector(".td-post-content > p")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.meta.get("article:published_time"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")



from typing import List

import lxml.html
import more_itertools
import pytest
import requests
from lxml.etree import XPath

from fundus import PublisherCollection
from fundus.publishers import Publisher, PublisherGroup
from fundus.scraping.session import _default_header

_language_code_selector = XPath("//table[contains(@class, 'wikitable') and @id='Table'] //td[@id] / @id")


def get_two_letter_code() -> List[str]:
    wiki_page = requests.get("https://en.wikipedia.org/wiki/List_of_ISO_639_language_codes", headers=_default_header)
    two_letter_codes: List[str] = _language_code_selector(lxml.html.document_fromstring(wiki_page.content))
    return two_letter_codes


language_codes = get_two_letter_code()


class TestPublisherCollection:
    @pytest.mark.parametrize(
        "region",
        [pytest.param(group, id=group.__name__) for group in PublisherCollection.get_subgroup_mapping().values()],
    )
    def test_default_language(self, region: PublisherGroup):
        assert hasattr(region, "default_language"), f"Region {region.__name__!r} has no default language set"

        default_language = getattr(region, "default_language")

        assert (
            default_language in language_codes
        ), f"Default language {default_language!r} isn't a ISO 639 language code"

    @pytest.mark.parametrize(
        "publisher", [pytest.param(publisher, id=publisher.__name__) for publisher in PublisherCollection]
    )
    def test_source_languages(self, publisher: Publisher):
        for source in more_itertools.flatten(publisher.source_mapping.values()):
            assert source.languages.issubset(language_codes)



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class ExpressParser(ParserProxy):
    class V1(BaseParser):
        _summary_selector = CSSSelector("article > header > h3")
        # only relevant for live-tickers
        _subheadline_selector = CSSSelector("div.live-events h3")
        _paragraph_selector = CSSSelector(
            "article div.text-description:not(.dont-miss) > p, div.live-events div.live-events__entry-text > p"
        )

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.meta.get("article:published_time"))

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.meta.get("author"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("article:tag"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=CSSSelector("div[role=main] article"),
                image_selector=CSSSelector("div.photo img"),
                caption_selector=XPath("./ancestor::div[contains(@class, 'photo')]/span[@class='newsCaption']/text()"),
                author_selector=XPath(
                    "./ancestor::div[contains(@class, 'photo')]/span[@class='newsCaption']/span[@class='caption']"
                ),
                size_pattern=re.compile(r"/(?P<width>[0-9]+)x(?P<height>[0-9]+)?/"),
            )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap

from ..shared import EuronewsParser
from .le_figaro import LeFigaroParser
from .le_monde import LeMondeParser
from .les_echos import LesEchosParser


class FR(metaclass=PublisherGroup):
    default_language = "fr"

    LeMonde = Publisher(
        name="Le Monde",
        domain="https://www.lemonde.fr/",
        parser=LeMondeParser,
        sources=[
            Sitemap("https://www.lemonde.fr/sitemap_index.xml"),
            NewsMap("https://www.lemonde.fr/sitemap_news.xml"),
        ],
    )

    EuronewsFR = Publisher(
        name="Euronews (FR)",
        domain="https://fr.euronews.com/",
        parser=EuronewsParser,
        sources=[
            Sitemap("https://fr.euronews.com/sitemaps/fr/articles.xml"),
            NewsMap("https://fr.euronews.com/sitemaps/fr/latest-news.xml"),
        ],
    )

    LeFigaro = Publisher(
        name="Le Figaro",
        domain="https://www.lefigaro.fr/",
        parser=LeFigaroParser,
        sources=[
            Sitemap("https://sitemaps.lefigaro.fr/lefigaro.fr/articles.xml"),
            NewsMap("https://www.lefigaro.fr/sitemap_news.xml"),
            RSSFeed("https://www.lefigaro.fr/rss/figaro_actualites.xml"),
            RSSFeed("https://www.lefigaro.fr/rss/figaro_flash-actu.xml"),
        ],
    )

    LesEchos = Publisher(
        name="Les Échos",
        domain="https://www.lesechos.fr/",
        parser=LesEchosParser,
        sources=[
            Sitemap("https://sitemap.lesechos.fr/sitemap_index.xml", reverse=True),
            NewsMap("https://www.lesechos.fr/sitemap_news.xml"),
        ],
        deprecated=True,
    )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.publishers.tw.taipei_times import TaipeiTimesParser
from fundus.scraping.url import NewsMap, Sitemap


class TW(metaclass=PublisherGroup):
    default_language = "tw"

    TaipeiTimes = Publisher(
        name="Taipei Times",
        domain="https://www.taipeitimes.com/",
        parser=TaipeiTimesParser,
        sources=[
            Sitemap("https://www.taipeitimes.com/sitemapIndex.xml", languages={"en"}),
            NewsMap("https://www.taipeitimes.com/sitemap/sitemap.xml", languages={"en"}),
        ],
    )



import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, ParserProxy, attribute
from fundus.parser.data import Image
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class BusinessInsiderDEParser(ParserProxy):
    class V1(BaseParser):
        _summary_selector = CSSSelector("article div.bi-bulletpoints li, article div.bi-bulletpoints > p")
        _subheadline_selector = CSSSelector("article > div > h2, article > div > h3")
        _paragraph_selector = XPath(
            """
            //article
            //div[
                contains(@class, 'article-body') 
                or contains(@class, 'piano-article')]
            /p[
                not(ancestor::*[@class='bi-bulletpoints']
                    or mark[@class='has-inline-color has-cyan-bluish-gray-color']
                    or @class='has-text-align-right')]
            """
        )

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("keywords")) or generic_topic_parsing(
                self.precomputed.ld.bf_search("keywords")
            )

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//img[not(contains(@class, 'size-thumbnail-square'))]"),
            )



from datetime import date, datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    image_extraction,
)


class TheInterceptParser(ParserProxy):
    class V1(BaseParser):
        # this date isn't exact nor confirmed, because TheIntercept isn't included an CC-NEWS and the
        # Publisher Coverage action didn't capture the layout change.
        VALID_UNTIL = date(2024, 2, 1)
        _summary_selector: XPath = CSSSelector("h2.Post-excerpt")
        _paragraph_selector: XPath = CSSSelector(
            "div.PostContent > div > p:not(p.caption):not(p.PhotoGrid-description)"
        )
        _subheadline_selector: XPath = CSSSelector("div.PostContent > div > h2")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                # The Intercept uses `p` tags for the article's paragraphs, image captions and photo grid descriptions.
                # Since we are only interested in the article's paragraphs,
                # we exclude the other elements from the paragraph selector.
                # Example article: https://theintercept.com/2023/04/01/israel-palestine-apartheid-settlements/
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.xpath_search("NewsArticle/author"))

        @attribute
        def publishing_date(self) -> Optional[datetime]:
            return generic_date_parsing(self.precomputed.ld.xpath_search("NewsArticle/datePublished", scalar=True))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.xpath_search("NewsArticle/headline", scalar=True)

        @attribute
        def topics(self) -> List[str]:
            # The Intercept specifies the article's topics, including other metadata,
            # inside the "keywords" linked data indicated by a "Subject: " prefix.
            # Example keywords: ["Day: Saturday", ..., "Subject: World", ...]
            keywords: List[str] = self.precomputed.ld.xpath_search("NewsArticle/keywords")
            return [keyword[9:] for keyword in keywords if keyword.startswith("Subject: ")]

    class V1_1(V1):
        VALID_UNTIL = date.today()
        _summary_selector = XPath(
            "//p[@class='post__excerpt'] | //h2[preceding-sibling::h1[contains(@class, 'post__title')]]"
        )
        _paragraph_selector = CSSSelector("div.entry-content > div.entry-content__content > p, blockquote > p")
        _subheadline_selector = CSSSelector("div.entry-content > div.entry-content__content > h2")

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath(
                    "//img[(string-length(@alt) > 0 and not(contains(@class, 'attachment') or contains(@class, ':hidden'))) or @loading='eager']|//figure//img"
                ),
                caption_selector=XPath(
                    "(./parent::article//div[contains(@class, 'image__caption')]/span[not(@class)])[1]|"
                    "./ancestor::figure//figcaption/span[@class='photo__caption']"
                ),
                author_selector=XPath(
                    "(./parent::article//div[contains(@class, 'image__caption')]/span)[last()]|"
                    "./ancestor::figure//figcaption/span[@class='photo__credit']"
                ),
            )



import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, ParserProxy, attribute
from fundus.parser.data import Image
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class NettavisenParser(ParserProxy):
    class V1(BaseParser):
        _bloat_pattern: str = "Les også:"

        _summary_selector = CSSSelector("p.lead-text")
        _subheadline_selector = CSSSelector("div.article-body > h2")
        _paragraph_selector = XPath(
            f"//div[contains(@class,'article-body')] /p[not(re:test(string(), '{_bloat_pattern}'))]",
            namespaces={"re": "http://exslt.org/regular-expressions"},
        )

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                subheadline_selector=self._subheadline_selector,
                summary_selector=self._summary_selector,
            )

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.meta.get("article:published_time"))

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.meta.get("article:author"))

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("article:tag"))

        @attribute
        def images(self) -> List[Image]:
            author_pattern = r"(Foto:\s*).*$"
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//brick-image-v3 | //img"),
                caption_selector=XPath("./ancestor::div[contains(@class, 'image')]//span[1]"),
                author_selector=XPath(
                    f"re:match(./ancestor::div[contains(@class, 'image')]//span[2], '{author_pattern}')",
                    namespaces={"re": "http://exslt.org/regular-expressions"},
                ),
            )



import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_date_parsing,
    image_extraction,
)


class PostillonParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = CSSSelector("div[id=post-body] p")
        _postscript_selector = CSSSelector("div[id=post-body] > span")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def authors(self) -> List[str]:
            postscript = self._postscript_selector(self.precomputed.doc)
            if not postscript:
                return []
            author_line = postscript[0].text_content().split(";")[0]
            return [a.strip() for a in author_line.split(",")]

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//div[@class='entry-content-wrap flex-col']//img"),
            )



import datetime
import re
from typing import List, Optional

from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    image_extraction,
)


class ThePortugalNewsParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = XPath("//div[@class='article-body']//p[string-length(text())>1]")
        _subheadline_selector = XPath("//div[@class='article-body']/p/b[not(u)]")
        _summary_selector = XPath("//div[@class='fs-4 font-semibold mb-3']")

        _author_selector = XPath("//div[@class='col-lg-10 order-lg-1']/p//text()")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                subheadline_selector=self._subheadline_selector,
                summary_selector=self._summary_selector,
            )

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def authors(self) -> List[str]:
            author_objects = self._author_selector(self.precomputed.doc)
            if author_objects and (author := re.search(r"(?i)by\s*(?P<authors>.*),[\r\sr\n]*in", author_objects[0])):
                return generic_author_parsing(author.group("authors"))
            return []

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                author_selector=re.compile(r"(?i)credits:\s*(?P<credits>.*)"),
            )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap

from .asahi_shimbun import AsahiShimbunParser
from .mainichi_shimbun import MainichiShimbunParser
from .nikkan_geadai import NikkanGeadaiParser
from .nikkei import NikkeiParser
from .sankei_shimbun import SankeiShimbunParser
from .the_japan_news import TheJapanNewsParser
from .tokyo_chunichi_shimbun import TokyoChunichiShimbunParser
from .yomiuri_shimbun import YomiuriShimbunParser


class JP(metaclass=PublisherGroup):
    default_language = "ja"

    TheJapanNews = Publisher(
        name="The Japan News",
        domain="https://japannews.yomiuri.co.jp/",
        parser=TheJapanNewsParser,
        sources=[
            Sitemap(
                "https://japannews.yomiuri.co.jp/sitemap.xml",
                sitemap_filter=regex_filter(r"(sitemap-news|sitemap-root|category)"),
                languages={"en"},
            ),
            NewsMap(
                "https://japannews.yomiuri.co.jp/sitemap-news.xml",
                languages={"en"},
            ),
        ],
    )

    YomiuriShimbun = Publisher(
        name="Yomiuri Shimbun",
        domain="https://www.yomiuri.co.jp/",
        parser=YomiuriShimbunParser,
        sources=[
            Sitemap(
                "https://www.yomiuri.co.jp/sitemap.xml",
                sitemap_filter=regex_filter("sitemap-news-latest"),
            ),
            NewsMap("https://www.yomiuri.co.jp/sitemap-news-latest.xml"),
        ],
    )

    AsahiShimbun = Publisher(
        name="Asahi Shimbun",
        domain="https://www.asahi.com/",
        parser=AsahiShimbunParser,
        sources=[NewsMap("https://www.asahi.com/sitemap.xml")],
    )

    TokyoShimbun = Publisher(
        name="Tokyo Shimbun",
        domain="https://www.tokyo-np.co.jp/",
        parser=TokyoChunichiShimbunParser,
        sources=[NewsMap("https://www.tokyo-np.co.jp/sitemap.xml")],
    )

    ChunichiShimbun = Publisher(
        name="Chunichi Shimbun",
        domain="https://www.chunichi.co.jp/",
        parser=TokyoChunichiShimbunParser,
        sources=[NewsMap("https://www.chunichi.co.jp/sitemap.xml")],
    )

    MainichiShimbun = Publisher(
        name="Mainichi Shimbun",
        domain="https://mainichi.jp/",
        parser=MainichiShimbunParser,
        sources=[
            RSSFeed("https://mainichi.jp/rss/etc/mainichi-flash.rss"),
        ],
    )

    Nikkei = Publisher(
        name="The Nikkei",
        domain="https://www.nikkei.com/",
        parser=NikkeiParser,
        sources=[
            NewsMap(
                "https://www.nikkei.com/sitemap.xml",
                sitemap_filter=inverse(regex_filter(r"[a-z]*\.sitemap\.xml$")),
            )
        ],
        deprecated=True,
    )

    SankeiShimbun = Publisher(
        name="Sankei Shimbun",
        domain="https://www.sankei.com/",
        parser=SankeiShimbunParser,
        sources=[
            # The Google sitemap https://www.sankei.com/feeds/google-sitemapindex/ is included here as well
            NewsMap("https://www.sankei.com/feeds/sitemapindex-category/?outputType=xml"),
        ],
    )

    NikkanGeadai = Publisher(
        name="Nikkan Geadai",
        domain="https://www.nikkan-gendai.com/",
        parser=NikkanGeadaiParser,
        sources=[
            Sitemap(
                "https://www.nikkan-gendai.com/sitemap.xml",
                reverse=True,
                sitemap_filter=inverse(regex_filter(r"type=articles")),
            )
        ],
    )



import datetime
import re
from typing import List, Optional

from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class DailyMaverickParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = XPath(
            "//div[contains(@class,' mode-content article-content ')]"
            "//p[(span or a and not(b)) or (text() and not(re:test(string(.), '^([A-Z ]+|Read more:.*)$')))]",
            namespaces={"re": "http://exslt.org/regular-expressions"},
        )
        _summary_selector = XPath("(//h2[@class='first-paragraph'])[1]")
        _subheadline_selector = XPath(
            "//div[contains(@class,' mode-content article-content ')]//h4 | "
            "//div[contains(@class,' mode-content article-content ')]//p[re:test(string(.), '^[A-Z ]+$')]",
            namespaces={"re": "http://exslt.org/regular-expressions"},
        )

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def topics(self) -> List[str]:
            return [
                t for t in generic_topic_parsing(self.precomputed.ld.bf_search("keywords")) if t not in self.authors()
            ]

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath("//h1"),
                image_selector=XPath(
                    "//div[contains(@class, 'article-body')]/img | //div[contains(@class, 'wp-caption')]/img"
                ),
                caption_selector=XPath(
                    "./self::img[contains(@class, 'header-image')]/ancestor::div[contains(@class, 'article-body')]//div[@class='image-caption'] |"
                    "./ancestor::div[contains(@class, 'wp-caption')]//p[@class='wp-caption-text']"
                ),
                author_selector=re.compile(r"(?i)\(photo:(?P<credits>.+)\)"),
            )



import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class FoxNewsParser(ParserProxy):
    class V1(BaseParser):
        _summary_selector = CSSSelector(".article-meta > h2")
        _paragraph_selector = XPath(
            "(//div[@class='article-body'] | //div[@class='article-body']/div[contains(@class, 'paywall')]) "
            "/p[not(child::script) and text()]"
        )

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.meta.get("dc.creator"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("classification-tags"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//article//picture//img[not(@*[starts-with(name(), 'data-v-')])]"),
                caption_selector=XPath("(./ancestor::div[@class='image-ct inline']//div[@class='caption']/p/span)[1]"),
                author_selector=XPath(
                    "(./ancestor::div[@class='image-ct inline']//div[@class='caption']/p/span)[last()]"
                ),
            )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import Sitemap

from .the_portugal_news import ThePortugalNewsParser


class PT(metaclass=PublisherGroup):
    default_language = "pt"

    ThePortugalNews = Publisher(
        name="The Portugal News",
        domain="https://www.theportugalnews.com/",
        parser=ThePortugalNewsParser,
        sources=[
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-en.xml")),
                languages={"en"},
            ),
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-de.xml")),
                languages={"de"},
            ),
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-nl.xml")),
                languages={"nl"},
            ),
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-fr.xml")),
                languages={"fr"},
            ),
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-es.xml")),
                languages={"es"},
            ),
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-it.xml")),
                languages={"it"},
            ),
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-se.xml")),
                languages={"se"},
            ),
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-ru.xml")),
                languages={"ru"},
            ),
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-zh.xml")),
                languages={"zh"},
            ),
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-tr.xml")),
                languages={"tr"},
            ),
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-pt.xml")),
            ),
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-ar.xml")),
                languages={"ar"},
            ),
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-he.xml")),
                languages={"he"},
            ),
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-pl.xml")),
                languages={"pl"},
            ),
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-fi.xml")),
                languages={"fi"},
            ),
            Sitemap(
                "https://www.theportugalnews.com/sitemap-news.xml",
                sitemap_filter=inverse(regex_filter("news-br.xml")),
                languages={"br"},
            ),
        ],
    )



from datetime import datetime
from textwrap import TextWrapper, dedent
from typing import Any, Dict, List, Mapping, Optional

import langdetect
import lxml.html
from colorama import Fore, Style

from fundus.logging import create_logger
from fundus.parser import ArticleBody, Image
from fundus.scraping.html import HTML
from fundus.utils.serialization import JSONVal, is_jsonable

logger = create_logger(__name__)


class AttributeView:
    def __init__(self, key: str, extraction: Mapping[str, Any]):
        self.ref = extraction
        self.key = key

    def __get__(self, instance: object, owner: type):
        return self.ref[self.key]

    def __set__(self, obj, value):
        # For now, this is read-only
        raise AttributeError("attribute is read only")


class Article:
    __extraction__: Mapping[str, Any] = {}

    def __init__(self, *, html: HTML, exception: Optional[Exception] = None, **extraction: Any) -> None:
        self.html = html
        self.exception = exception
        self.__extraction__ = extraction

        # create descriptors for attributes that aren't pre-defined as properties.
        for attribute in extraction.keys():
            if not hasattr(self, attribute):
                setattr(self, attribute, AttributeView(attribute, self.__extraction__))

    @property
    def title(self) -> Optional[str]:
        return self.__extraction__.get("title")

    @property
    def body(self) -> Optional[ArticleBody]:
        return self.__extraction__.get("body")

    @property
    def authors(self) -> List[str]:
        return self.__extraction__.get("authors", [])

    @property
    def publishing_date(self) -> Optional[datetime]:
        return self.__extraction__.get("publishing_date")

    @property
    def topics(self) -> List[str]:
        return self.__extraction__.get("topics", [])

    @property
    def free_access(self) -> bool:
        return self.__extraction__.get("free_access", False)

    @property
    def images(self) -> List[Image]:
        return self.__extraction__.get("images", [])

    @property
    def publisher(self) -> str:
        return self.html.source_info.publisher

    def __getattribute__(self, item: str):
        if (attribute := object.__getattribute__(self, item)) and hasattr(attribute, "__get__"):
            return attribute.__get__(self, type(self))
        return attribute

    def __setattr__(self, key: str, value: object):
        if hasattr(self, key):
            # we can't use getattr here, because it would invoke __get__, so unfortunately no default value
            attribute = object.__getattribute__(self, key)
            if hasattr(attribute, "__set__"):
                attribute.__set__(key, value)
                return
        object.__setattr__(self, key, value)

    def __getattr__(self, item: str):
        raise AttributeError(f"{type(self).__name__!r} object has no attribute {str(item)!r}")

    @property
    def plaintext(self) -> Optional[str]:
        return str(self.body) or None if not isinstance(self.body, Exception) else None

    @property
    def lang(self) -> Optional[str]:
        language: Optional[str] = None

        if self.plaintext:
            try:
                language = langdetect.detect(self.plaintext)
            except langdetect.LangDetectException:
                logger.debug(f"Unable to detect language for article {self.html.responded_url!r}")

        # use @lang attribute of <html> tag as fallback
        if not language or language == langdetect.detector_factory.Detector.UNKNOWN_LANG:
            language = lxml.html.fromstring(self.html.content).get("lang")
            if language and "-" in language:
                language = language.split("-")[0]

        return language

    def to_json(self, *attributes: str) -> Dict[str, JSONVal]:
        """Converts article object into a JSON serializable dictionary.

        One can specify which attributes should be included by passing attribute names as parameters.
        Default: title, plaintext, authors, publishing_date, topics, free_access + unvalidated attributes

        Args:
            *attributes: The attributes to serialize. Default: see docstring.

        Returns:
            A json serializable dictionary
        """

        # default value for attributes
        if not attributes:
            attributes = tuple(set(self.__extraction__.keys()) - {"meta", "ld"})

        def serialize(v: Any) -> JSONVal:
            if hasattr(v, "serialize"):
                return v.serialize()  # type: ignore[no-any-return]
            elif isinstance(v, datetime):
                return str(v)
            elif not is_jsonable(v):
                raise TypeError(f"Attribute {attribute!r} of type {type(v)!r} is not JSON serializable")
            return v  # type: ignore[no-any-return]

        serialization: Dict[str, JSONVal] = {}
        for attribute in attributes:
            if not hasattr(self, attribute):
                continue
            value = getattr(self, attribute)

            if isinstance(value, list):
                serialization[attribute] = [serialize(item) for item in value]
            else:
                serialization[attribute] = serialize(value)

        return serialization

    def __str__(self):
        # the subsequent indent here is a bit wacky, but textwrapper.dedent won't work with tabs, so we have to use
        # whitespaces instead.
        title_wrapper = TextWrapper(width=80, max_lines=1, initial_indent="")
        text_wrapper = TextWrapper(width=80, max_lines=2, initial_indent="", subsequent_indent="          ")
        wrapped_title = title_wrapper.fill(
            f"{Fore.RED}--missing title--{Style.RESET_ALL}" if self.title is None else self.title.strip()
        )
        wrapped_plaintext = text_wrapper.fill(
            f"{Fore.RED}--missing plaintext--{Style.RESET_ALL}" if self.plaintext is None else self.plaintext.strip()
        )

        image_text = (
            f" including {len(self.images)} image(s)" if self.images and not isinstance(self.images, Exception) else ""
        )

        text = (
            f"Fundus-Article{image_text}:"
            f'\n- Title: "{wrapped_title}"'
            f'\n- Text:  "{wrapped_plaintext}"'
            f"\n- URL:    {self.html.requested_url}"
            f"\n- From:   {self.publisher}"
            f'{" (" + self.publishing_date.strftime("%Y-%m-%d %H:%M") + ")" if self.publishing_date else ""}'
        )

        return dedent(text)



import re
from typing import Callable, Dict, Literal, Optional, Pattern, TypeVar, Union, overload

_T = TypeVar("_T")


@overload
def _get_match_dict(pattern: Pattern[str], string: str, conversion: Callable[[str], _T]) -> Dict[str, _T]:
    ...


@overload
def _get_match_dict(
    pattern: Pattern[str], string: str, conversion: Callable[[str], _T], keep_none: Literal[True]
) -> Dict[str, Optional[_T]]:
    ...


@overload
def _get_match_dict(pattern: Pattern[str], string: str) -> Dict[str, str]:
    ...


@overload
def _get_match_dict(pattern: Pattern[str], string: str, keep_none: Literal[True]) -> Dict[str, Optional[str]]:
    ...


def _get_match_dict(  # type: ignore[misc]
    pattern: Pattern[str], string: str, conversion: Optional[Callable[[str], _T]] = None, keep_none: bool = False
) -> Dict[str, Union[str, _T, None]]:
    matches = {}
    for match in re.finditer(pattern, string):
        match_dict = match.groupdict()
        for key, value in match_dict.items():
            if value is not None:
                matches[key] = conversion(value) if conversion is not None else value
            elif keep_none:
                matches[key] = match[key] or value
    return matches



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap

from .dagbladet import DagbladetParser
from .nettavisen import NettavisenParser
from .nrk import NRKParser
from .verdensgang import VerdensGangParser


class NO(metaclass=PublisherGroup):
    default_language = "no"

    VerdensGang = Publisher(
        name="Verdens Gang",
        domain="https://www.vg.no/",
        parser=VerdensGangParser,
        sources=[
            Sitemap(
                "https://www.vg.no/sitemap.xml",
                sitemap_filter=inverse(regex_filter(r"vg\.no\/sitemaps/\d{4}\-\d{2}-articles.xml")),
                reverse=True,
            ),
            NewsMap("https://www.vg.no/sitemap/files/articles-48hrs.xml"),
        ],
    )

    Dagbladet = Publisher(
        name="Dagbladet",
        domain="https://www.db.no/",
        parser=DagbladetParser,
        sources=[
            Sitemap("https://www.dagbladet.no/app/jw-api-proxy/sitemap/index/dagbladet.xml", reverse=True),
            NewsMap("https://www.dagbladet.no/sitemap"),
        ],
    )

    Nettavisen = Publisher(
        name="Nettavisen",
        domain="https://www.nettavisen.no/",
        parser=NettavisenParser,
        sources=[
            RSSFeed("https://www.nettavisen.no/service/rich-rss"),
        ],
    )

    NRK = Publisher(
        name="Norsk rikskringkasting",
        domain="https://www.nrk.no/",
        parser=NRKParser,
        sources=[
            RSSFeed("https://www.nrk.no/norge/toppsaker.rss"),
            RSSFeed("https://www.nrk.no/urix/toppsaker.rss"),
            RSSFeed("https://www.nrk.no/sport/toppsaker.rss"),
            RSSFeed("https://www.nrk.no/kultur/toppsaker.rss"),
            RSSFeed("https://www.nrk.no/livsstil/toppsaker.rss"),
            RSSFeed("https://www.nrk.no/viten/toppsaker.rss"),
            RSSFeed("https://www.nrk.no/dokumentar/toppsaker.rss"),
            Sitemap("https://www.nrk.no/sitemap.xml"),
        ],
    )



from datetime import date, datetime
from typing import List, Optional

import lxml.html
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, ParserProxy, attribute, function
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
)


class ReutersParser(ParserProxy):
    class V1(BaseParser):
        VALID_UNTIL = date(2024, 1, 1)
        _paragraph_selector = XPath("(//p[starts-with(@data-testid, 'paragraph')])[position() > 1]")
        _summary_selector = XPath("(//p[starts-with(@data-testid, 'paragraph')])[1]")
        _subheadline_selector = XPath("//div[contains(@class, 'article-body')] /h2[@data-testid='Heading']")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                paragraph_selector=self._paragraph_selector,
                subheadline_selector=self._subheadline_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            # Reuters also has authors listed in the ld+json.
            # But there, the author names are listed at "byline" instead of "name".
            # This is not supported by the generic author parsing.
            return generic_author_parsing(self.precomputed.meta.get("article:author"))

        @attribute
        def publishing_date(self) -> Optional[datetime]:
            return generic_date_parsing(self.precomputed.ld.xpath_search("NewsArticle/datePublished", scalar=True))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.xpath_search("NewsArticle/headline", scalar=True)

        @attribute
        def topics(self) -> List[str]:
            # Reuters does not have very meaningful topics in the "keywords" meta.
            # Example: ['BLR', 'EGS', 'SOC', 'SOCC', 'SPO', ...]
            # But interesting topics from the meta may be found in these fields:
            #   - article:section                       ("Aerospace & Defense")
            #   - analyticsAttributes.topicChannel      ("Business")
            #   - analyticsAttributes.topicSubChannel   ("Aerospace & Defense")
            #   - DCSext.ChannelList                    ("Business;Asia Pacific;World")
            topics: List[Optional[str]] = [
                self.precomputed.meta.get("article:section"),
                self.precomputed.meta.get("analyticsAttributes.topicChannel"),
                self.precomputed.meta.get("analyticsAttributes.topicSubChannel"),
            ]
            topics.extend(generic_topic_parsing(self.precomputed.meta.get("DCSext.ChannelList"), delimiter=";"))

            # Remove empty topics and duplicates deterministically
            processed_topics = list(dict.fromkeys(topic for topic in topics if topic))
            return processed_topics

    class V1_1(V1):
        VALID_UNTIL = date.today()

        # TODO: at the end of sports related articles like
        #  https://www.reuters.com/sports/basketball/hot-shooting-suns-wear-down-raptors-2024-03-08/
        #  there is this `--Field Level Media` bloat line
        _paragraph_selector = XPath("(//div[starts-with(@data-testid, 'paragraph')])[position() > 1]")
        _summary_selector = XPath("(//div[starts-with(@data-testid, 'paragraph')])[1]")

        _new_tab_span_selector = XPath(
            "//div[starts-with(@data-testid, 'paragraph')] //span[contains(text(), 'opens new tab')]"
        )

        @function(priority=1)
        def _remove_new_tab_span(self) -> None:
            span: lxml.html.HtmlElement
            for span in self._new_tab_span_selector(self.precomputed.doc):
                span.drop_tree()



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.url import RSSFeed, Sitemap

from .people import PeopleParser


class CN(metaclass=PublisherGroup):
    default_language = "zh"

    People = Publisher(
        name="People",
        domain="http://www.people.com.cn",
        parser=PeopleParser,
        sources=[
            RSSFeed("http://www.people.com.cn/rss/politics.xml"),
            RSSFeed("http://www.people.com.cn/rss/society.xml"),
            RSSFeed("http://www.people.com.cn/rss/legal.xml"),
            RSSFeed("http://www.people.com.cn/rss/world.xml"),
            RSSFeed("http://www.people.com.cn/rss/haixia.xml"),
            RSSFeed("http://www.people.com.cn/rss/military.xml"),
            RSSFeed("http://www.people.com.cn/rss/ywkx.xml"),
            Sitemap("http://www.people.cn/sitemap_index.xml"),
        ],
    )



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class LRTParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = XPath(
            "//div[@class='article-content js-text-selection']/"
            "p[not((strong and not(text())) or @class='text-lead')]"
        )
        _summary_selector = CSSSelector("p.text-lead")
        _subheadline_selector = XPath(
            "//div[@class='article-content js-text-selection']/" "p[strong and not(@class='text-lead') and not(text())]"
        )

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.meta.get("lrt_authors"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.ld.bf_search("keywords"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=CSSSelector("article"),
                image_selector=CSSSelector("div.media-block img"),
                caption_selector=XPath(
                    "./ancestor::div[contains(@class, 'media-block')]//div[contains(@class, 'description')]"
                ),
                author_selector=re.compile(r"/\s*(?P<credits>.*).$"),
                relative_urls=True,
            )



import datetime
import re
from typing import List, Optional

from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class MalayMailParser(ParserProxy):
    class V1(BaseParser):
        # Use negated _subheadline_selector as _paragraph_selector to ensure no double parsing
        _paragraph_selector = XPath("//div[@class='article-body']/p[text() or not(b)]")
        _subheadline_selector = XPath("//div[@class='article-body']/p[not(text()) and b]")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                subheadline_selector=self._subheadline_selector,
            )

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("keywords"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//div[contains(@class, 'image')]//img"),
                caption_selector=XPath("(./ancestor::div[contains(@class, 'image')])[1]//div[@class='image-caption']"),
                author_selector=re.compile(r"\s*—\s*(?P<credits>.*)$"),
            )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import Sitemap

from .jyllands_posten import JyllandsPostenParser


class DK(metaclass=PublisherGroup):
    default_language = "da"

    JyllandsPosten = Publisher(
        name="Jyllands Posten",
        domain="https://jyllands-posten.dk/",
        parser=JyllandsPostenParser,
        sources=[
            Sitemap(
                "https://jyllands-posten.dk/sitemapindex.xml",
                reverse=True,
            ),
        ],
        url_filter=inverse(regex_filter(r"/ECE")),
    )



import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class EveningStandardParser(ParserProxy):
    class V1(BaseParser):
        VALID_UNTIL = datetime.date(2024, 6, 30)
        _paragraph_selector = CSSSelector("div.sc-bkSUFG.bdkDcZ")
        _summary_selector = CSSSelector("div.sc-wkolL.dWZJhQ")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                summary_selector=self._summary_selector,
            )

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("keywords"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//picture[not(ancestor::a)] /img"),
                upper_boundary_selector=CSSSelector("article"),
                caption_selector=XPath(
                    "./ancestor::div[count(div)=3 and position() <= 2]/div[2] |"
                    "./ancestor::div[picture and count(div)=2][1]/div[1]"
                ),
                author_selector=XPath(
                    "./ancestor::div[count(div)=3 and position() <= 2]/div[3] |"
                    "./ancestor::div[picture and count(div)=2][1]/div[2]"
                ),
                lower_boundary_selector=CSSSelector("div#piano-reg-wall"),
            )

    class V1_1(V1):
        VALID_UNTIL = datetime.date.today()
        _summary_selector = CSSSelector("div.sc-jgyXzG")
        _subheadline_selector = CSSSelector("div#main div.sc-dFfFtc > h3")
        _paragraph_selector = CSSSelector("div#main > div.sc-gEvEer p")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class SportSchauParser(ParserProxy):
    class V1(BaseParser):
        _summary_selector = CSSSelector(
            "p[class='textabsatz columns twelve  m-ten  m-offset-one l-eight l-offset-two']" " > strong"
        )
        _paragraph_selector = CSSSelector("article >p.textabsatz:not(p.textabsatz:nth-of-type(1))")
        _subheadline_selector = CSSSelector("article >h2")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.meta.get("author"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(
                self.precomputed.meta.get(
                    "date",
                )
            )

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("keywords"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//article//picture[not(contains(@class,'--list'))]//img"),
                lower_boundary_selector=XPath("//div[contains(@class, 'back-to-top')]"),
                alt_selector=XPath("./@title"),
                author_selector=re.compile(r"\|(?P<credits>.+)"),
                caption_selector=XPath(
                    "./ancestor::div[contains(@class, 'absatzbild ')]/div[@class='absatzbild__info']"
                ),
                size_pattern=re.compile(r"/[\dx]+-(?P<width>[0-9]+)/"),
            )



from typing import Dict, Iterator, List, Literal, Optional, Type

import more_itertools

from fundus.logging import create_logger
from fundus.parser import ParserProxy
from fundus.publishers.base_objects import Publisher
from fundus.scraping.article import Article
from fundus.scraping.delay import Delay
from fundus.scraping.filter import (
    ExtractionFilter,
    FilterResultWithMissingAttributes,
    URLFilter,
)
from fundus.scraping.html import CCNewsSource, HTMLSource, WebSource
from fundus.scraping.url import URLSource
from fundus.utils.events import __EVENTS__

logger = create_logger(__name__)


class BaseScraper:
    def __init__(self, *sources: HTMLSource, parser_mapping: Dict[str, ParserProxy]):
        self.sources = sources
        self.parser_mapping = parser_mapping

    def scrape(
        self,
        error_handling: Literal["suppress", "catch", "raise"],
        extraction_filter: Optional[ExtractionFilter] = None,
        url_filter: Optional[URLFilter] = None,
        language_filter: Optional[List[str]] = None,
    ) -> Iterator[Article]:
        for source in self.sources:
            for html in source.fetch(url_filter=url_filter):
                parser = self.parser_mapping[html.source_info.publisher]

                try:
                    extraction = parser(html.crawl_date).parse(html.content, error_handling)

                except Exception as error:
                    if error_handling == "raise":
                        error_message = f"Run into an error processing article {html.requested_url!r}"
                        logger.error(error_message)
                        error.args = (str(error) + "\n\n" + error_message,)
                        raise error
                    elif error_handling == "catch":
                        yield Article(html=html, exception=error)
                    elif error_handling == "suppress":
                        logger.info(f"Skipped article at {html.requested_url!r} because of: {error!r}")
                    else:
                        raise ValueError(f"Unknown value {error_handling!r} for parameter <error_handling>'")

                else:
                    if extraction_filter and (filter_result := extraction_filter(extraction)):
                        if isinstance(filter_result, FilterResultWithMissingAttributes):
                            logger.debug(
                                f"Skipped article at {html.requested_url!r} because attribute(s) "
                                f"{', '.join(filter_result.missing_attributes)!r} is(are) missing"
                            )
                        else:
                            logger.debug(f"Skipped article at {html.requested_url!r} because of extraction filter")
                    else:
                        article = Article(html=html, **extraction)
                        if language_filter and article.lang not in language_filter:
                            logger.debug(
                                f"Skipped article at {html.requested_url!r} because article language: "
                                f"{article.lang!r} is not in allowed languages: {language_filter!r}"
                            )
                        else:
                            yield article


class WebScraper(BaseScraper):
    def __init__(
        self,
        publisher: Publisher,
        restrict_sources_to: Optional[List[Type[URLSource]]] = None,
        delay: Optional[Delay] = None,
        ignore_robots: bool = False,
        ignore_crawl_delay: bool = False,
    ):
        if restrict_sources_to:
            url_sources = tuple(
                more_itertools.flatten(
                    publisher.source_mapping[source_type]
                    for source_type in restrict_sources_to
                    if source_type in publisher.source_mapping
                )
            )
        else:
            url_sources = tuple(more_itertools.flatten(publisher.source_mapping.values()))

        html_sources = [
            WebSource(
                url_source=url_source,
                publisher=publisher,
                request_header=publisher.request_header,
                delay=delay,
                url_filter=publisher.url_filter,
                query_parameters=publisher.query_parameter,
                ignore_robots=ignore_robots,
                ignore_crawl_delay=ignore_crawl_delay,
            )
            for url_source in url_sources
        ]
        parser_mapping: Dict[str, ParserProxy] = {publisher.name: publisher.parser}
        super().__init__(*html_sources, parser_mapping=parser_mapping)

        __EVENTS__.alias(publisher.name)


class CCNewsScraper(BaseScraper):
    def __init__(self, source: CCNewsSource):
        parser_mapping: Dict[str, ParserProxy] = {publisher.name: publisher.parser for publisher in source.publishers}
        super().__init__(source, parser_mapping=parser_mapping)



import datetime
import re
from typing import List, Optional

from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class MallorcaMagazinParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = XPath("//div[@id='post-text']//p")
        _subheadline_selector = XPath("//div[@id='post-text']//*[(self::h4 or self::h2) and not(@class)]")
        _summary_selector = XPath("//h2[@class='post-subtitle']")

        _topic_selector = XPath("//div[@class='post-tags']//li")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                subheadline_selector=self._subheadline_selector,
                summary_selector=self._summary_selector,
            )

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def topics(self) -> List[str]:
            return [node.text_content().strip() for node in self._topic_selector(self.precomputed.doc)]

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                image_selector=XPath("//figure//img|//div[@id='post-text']//p/img"),
                paragraph_selector=self._paragraph_selector,
                caption_selector=XPath(
                    "./ancestor::div[@class='col-sm-12']//p[@class='img-description'] | "
                    "./ancestor::figure//figcaption"
                ),
                author_selector=re.compile(r"\|(?P<credits>.+)"),
            )



import datetime
from typing import List, Optional

from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
)


class IsraelNachrichtenParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = XPath("//div[@id='main']/div[@class]/p[text()]")
        _summary_selector = XPath("//div[@id='main']/div[@class]/p/strong")
        _title_selector = XPath("//div[@id='main']/div[@class]/h1")

        _author_selector = XPath("//div[@id='main']/div[@class]/p/em")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            if authors := self._author_selector(self.precomputed.doc):
                return generic_author_parsing(authors[0].text_content())
            return []

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.meta.get("article:published_time"))

        @attribute
        def title(self) -> Optional[str]:
            return (
                self._title_selector(self.precomputed.doc)[0].text_content().strip()
                if self._title_selector(self.precomputed.doc)
                else None
            )



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from dataclasses import dataclass, fields
from functools import total_ordering
from itertools import chain
from typing import (
    Any,
    ClassVar,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    overload,
)
from urllib.parse import urljoin, urlparse

import lxml.etree
import lxml.html
import more_itertools
import validators
import xmltodict
from dict2xml import dict2xml
from lxml.etree import XPath, fromstring, tostring
from typing_extensions import Self, TypeAlias, deprecated

from fundus.utils.serialization import (
    DataclassSerializationMixin,
    JSONVal,
    replace_keys_in_nested_dict,
)

LDMappingValue: TypeAlias = Union[List[Dict[str, Any]], Dict[str, Any]]

_sentinel = object()

_T = TypeVar("_T")


class LinkedDataMapping:
    """
    This class is a @type -> LD mapping.
    Given LD:
        ld_1 = {
            @type: 'Article'
            ...
            ...
        }
    Article will be mapped to ld_1

    If there are multiple LDs with the same type, this type will map to a list of LDs.
    In this context an LD is represented as a python dict.
    """

    __UNKNOWN_TYPE__ = "UNKNOWN_TYPE"
    __xml_transformation_table__ = {":": "U003A", "*": "U002A", "@": "U0040"}

    def __init__(self, lds: Iterable[Dict[str, Any]] = ()):
        for ld in lds:
            if graph := ld.get("@graph"):
                for nested in graph:
                    self.add_ld(nested)
            else:
                self.add_ld(ld)
        self.__xml: Optional[lxml.etree._Element] = None

    def __getstate__(self):
        state = self.__dict__.copy()
        if self.__xml is not None:
            state["_LinkedDataMapping__xml"] = tostring(self.__xml)
        return state

    def __setstate__(self, state):
        if (xml_element := state.get("_LinkedDataMapping__xml")) is not None:
            state["_LinkedDataMapping__xml"] = fromstring(xml_element)
        self.__dict__ = state

    def serialize(self) -> Dict[str, Any]:
        return {attribute: value for attribute, value in self.__dict__.items() if "__" not in attribute}

    def _add(self, ld: Dict[str, JSONVal], ld_type: str) -> None:
        if value := self.__dict__.get(ld_type):
            if not isinstance(value, list):
                self.__dict__[ld_type] = [value]
            self.__dict__[ld_type].append(ld)
        else:
            self.__dict__[ld_type] = ld

    def add_ld(self, ld: Dict[str, Any], name: Optional[str] = None) -> None:
        if ld_type := (name or ld.get("@type")):
            if isinstance(ld_type, str):
                self._add(ld, ld_type)
            elif isinstance(ld_type, list):
                for t in ld_type:
                    self._add(ld, t)
            else:
                raise NotImplementedError(f"Unexpected LD type {type(ld_type)}")
        else:
            self._add(ld, self.__UNKNOWN_TYPE__)

    @deprecated("Use xpath_search() instead")
    def get_value_by_key_path(self, key_path: List[str], default: Any = None) -> Optional[Any]:
        """
        Works like get() except this one assumes a path is given as list of keys (str).
        I.e:
            key_path := ["mainEntity", "author"], default := {}
            results in self._ld_by_type.get("mainEntity").get("author")

        Whenever a key is missing or an empty value occurs down the path this funktion will immediately return
        <default>, but will not catch if not all values supports get()

        :param key_path: A list of keys in order forming a path to the desired value
        :param default: A default returned when either a key is missing or resulting in an empty/null value
        :return: The reached value or <default>
        """
        tmp = self.__dict__.copy()
        for key in key_path:
            if not (nxt := tmp.get(key)):
                return default
            tmp = nxt
        return tmp

    def __as_xml__(self) -> lxml.etree._Element:
        pattern = re.compile("|".join(map(re.escape, self.__xml_transformation_table__.keys())))

        def to_unicode_characters(text: str) -> str:
            return pattern.sub(lambda match: self.__xml_transformation_table__[match.group(0)], text)

        if self.__xml is None:
            xml = dict2xml(replace_keys_in_nested_dict({"linkedData": self.serialize()}, to_unicode_characters))
            self.__xml = lxml.etree.fromstring(xml)
        return self.__xml

    @overload
    def xpath_search(self, query: Union[XPath, str], scalar: Literal[False] = False) -> List[Any]:
        ...

    @overload
    def xpath_search(self, query: Union[XPath, str], scalar: Literal[True] = True) -> Optional[Any]:
        ...

    def xpath_search(self, query: Union[XPath, str], scalar: bool = False):
        """Search through LD using XPath expressions

        Internally, the content of the LinkedDataMapping is converted to XML and then
        evaluated with an XPath expression <query>.

        To search for keys including invalid XML characters, use Unicode representations instead:
        I.e. to search for the key "_16:9" write "//_16U003A9"
        For all available transformations see LinkedDataMapping.__xml_transformation_table__

        Note that values will be converted to strings, i.e. True -> 'True', 1 -> '1'

        Examples:
            LinkedDataMapping = {
                "b": {
                    "key": value1,
                }
                "c": {
                    "key": value2,
                }
            }

        LinkedDataMapping.xpath_search(XPath("//key"))
        >> [value1, value2]

        LinkedDataMapping.xpath_search(XPath("//b/key"))
        >> [value1]

        Args:
            query: A XPath expression either as string or XPath object.
            scalar: If True, return an optional "scalar" value and raise a ValueError if there are more
                than one result to return; if False, return a list of results. Defaults to False.

        Returns:
            An ordered list of search results or an optional "scalar" result
        """

        if isinstance(query, str):
            query = XPath(query)

        pattern = re.compile("|".join(map(re.escape, self.__xml_transformation_table__.values())))

        def node2string(n: lxml.etree._Element) -> str:
            return "".join(
                chunk
                for chunk in chain(
                    (n.text,),
                    chain(*((tostring(child, with_tail=False, encoding=str), child.tail) for child in n.getchildren())),
                )
                if chunk
            )

        reversed_table = {v: k for k, v in self.__xml_transformation_table__.items()}

        def to_original_characters(text: str) -> str:
            return pattern.sub(lambda match: reversed_table[match.group(0)], text)

        nodes = query(self.__as_xml__())

        results = {}

        for i, node in enumerate(nodes):
            xml = f"<result{i}>" + node2string(node) + f"</result{i}>"
            results.update(replace_keys_in_nested_dict(xmltodict.parse(xml), to_original_characters))

        values = list(filter(bool, results.values()))

        if scalar:
            if not values:
                return None
            elif len(values) == 1:
                return values.pop()
            else:
                raise ValueError(f"Got multiple values when expecting a single scalar value")
        else:
            return values

    def bf_search(self, key: str, depth: Optional[int] = None, default: Optional[_T] = None) -> Union[Any, _T]:
        """
        This is a classic BF search on the nested dicts representing the JSON-LD. <key> specifies the dict key to
        search, <depth> the depth level. If the depth level is set to None, this method will search through the whole
        LD. It is important to notice that this will only return the value of the first matched key.
        For more precise operations consider using get_by_key_path().

        I.e:

            considering the following LD:
                MainPage
                    @type
                    @content
                    BreadcrumbList
                        ...
                        ...
                    NewsArticle
                        datePublished: ...
                        authors: ...

            the contents of 'MainPage' count as depth 1.

            So
                breadth_first_search('authors') -> None,

            whereas

                breadth_first_search('@content') -> the value of key '@content'

            and

                breadth_first_search('authors', 2) -> the value of key 'authors'

        :param key: The dict key to search for
        :param depth: The searched depth, default None
        :return: The content of the first matched key or None
        """

        def search_recursive(nodes: Iterable[LDMappingValue], current_depth: int):
            if current_depth == depth:
                return _sentinel
            else:
                new: List[Dict[str, Any]] = []
                for node in nodes:
                    if isinstance(node, list):
                        new.extend(node)
                        continue
                    elif (value := node.get(key, _sentinel)) is not _sentinel:
                        return value

                    nested_dicts: Iterable[Dict[str, Any]] = filter(
                        lambda obj: isinstance(obj, dict), more_itertools.collapse(node.values(), base_type=dict)
                    )
                    new.extend(nested_dicts)

                if not new:
                    return _sentinel

                return search_recursive(new, current_depth + 1)

        result = search_recursive([self.__dict__], 0)

        if result == _sentinel:
            return default

        return result

    def __repr__(self):
        return f"LD containing {', '.join(content)!r}" if (content := self.__dict__.keys()) else "Empty LD"


class TextSequence(Sequence[str]):
    def __init__(self, texts: Iterable[str]):
        self._data: Tuple[str, ...] = tuple(texts)

    @overload
    def __getitem__(self, i: int) -> str:
        ...

    @overload
    def __getitem__(self, s: slice) -> "TextSequence":
        ...

    def __getitem__(self, i):
        return self._data[i] if isinstance(i, int) else type(self)(self._data[i])

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    def __repr__(self) -> str:
        return repr(self._data)

    def __str__(self) -> str:
        return "\n".join(self)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TextSequence):
            return NotImplemented
        return self._data == other._data


@dataclass
class TextSequenceTree(ABC):
    """Base class to traverse and build trees of TextSequence."""

    def as_text_sequence(self) -> TextSequence:
        texts = [text for tl in self.df_traversal() for text in tl]
        return TextSequence(texts)

    def text(self, join_on: str = "\n\n") -> str:
        return join_on.join(self.as_text_sequence())

    def df_traversal(self) -> Iterable[TextSequence]:
        def recursion(o: object):
            if isinstance(o, TextSequence):
                yield o
            elif isinstance(o, Collection):
                for el in o:
                    yield from el
            else:
                yield o

        for value in self:
            yield from recursion(value)

    @abstractmethod
    def serialize(self) -> Dict[str, Any]:
        pass

    @classmethod
    @abstractmethod
    def deserialize(cls, serialized: Dict[str, Any]) -> Self:
        pass

    def __iter__(self) -> Iterator[Any]:
        field_values = [getattr(self, f.name) for f in fields(self)]
        yield from field_values

    def __str__(self):
        return self.text()

    def __bool__(self) -> bool:
        return bool(self.as_text_sequence())


@dataclass
class ArticleSection(TextSequenceTree):
    headline: TextSequence
    paragraphs: TextSequence

    def serialize(self) -> Dict[str, Any]:
        return {
            "headline": list(self.headline),
            "paragraphs": list(self.paragraphs),
        }

    @classmethod
    def deserialize(cls, serialized: Dict[str, Any]) -> Self:
        return cls(headline=TextSequence(serialized["headline"]), paragraphs=TextSequence(serialized["paragraphs"]))

    def __bool__(self):
        return bool(self.paragraphs)


@dataclass
class ArticleBody(TextSequenceTree):
    summary: TextSequence
    sections: List[ArticleSection]

    def serialize(self) -> Dict[str, Any]:
        return {
            "summary": list(self.summary),
            "sections": [section.serialize() for section in self.sections],
        }

    @classmethod
    def deserialize(cls, serialized: Dict[str, Any]) -> Self:
        return cls(
            summary=TextSequence(serialized["summary"]),
            sections=[ArticleSection.deserialize(section) for section in serialized["sections"]],
        )

    def __bool__(self):
        return any(bool(section) for section in self.sections)


@total_ordering
@dataclass
class Dimension(DataclassSerializationMixin):
    width: int
    height: int

    def __mul__(self, other: Union[float, int]) -> "Dimension":
        if isinstance(other, int):
            return Dimension(self.width * other, self.height * other)
        elif isinstance(other, float):
            return Dimension(round(self.width * other), round(self.height * other))
        else:
            raise NotImplementedError(
                f"'*' is not defined between {type(self).__name__!r} and {type(other).__name__!r}"
            )

    def __rmul__(self, other: Union[float, int]) -> "Dimension":
        return self.__mul__(other)

    def __repr__(self) -> str:
        return f"{self.width}x{self.height or '...'}"

    def __lt__(self, other: "Dimension") -> bool:
        if isinstance(other, Dimension):
            if self.width != other.width:
                return self.width < other.width
            else:
                return self.height < other.height
        raise NotImplementedError(f"'<' is not defined between {type(self).__name__!r} and {type(other).__name__!r}")

    def __hash__(self) -> int:
        return hash((self.width, self.height))

    @classmethod
    def from_ratio(
        cls,
        width: Optional[float] = None,
        height: Optional[float] = None,
        ratio: Optional[float] = None,
    ) -> Optional["Dimension"]:
        if width and height:
            return cls(round(width), round(height))
        elif width is not None:
            return cls(round(width), round((width / ratio) if ratio else 0))
        elif height is not None:
            return cls(round((height * ratio) if ratio else 0), round(height))
        else:
            return None


def remove_query_parameters_from_url(url: str) -> str:
    if any(parameter_indicator in url for parameter_indicator in ("?", "#")):
        return urljoin(url, urlparse(url).path)
    return url


@total_ordering
@dataclass
class ImageVersion(DataclassSerializationMixin):
    __FILE_FORMATS__: ClassVar[List[str]] = ["png", "jpg", "jpeg", "webp"]

    url: str
    query_width: Optional[str] = None
    size: Optional[Dimension] = None
    type: Optional[str] = None

    def __post_init__(self):
        if not self.type:
            url_without_query = remove_query_parameters_from_url(self.url)
            self.type = self._parse_type(url_without_query)

    def _parse_type(self, url: str) -> Optional[str]:
        if (file_format := url.split(".")[-1]) in self.__FILE_FORMATS__:
            if file_format == "jpg":
                file_format = "jpeg"
            return "image/" + file_format
        return None

    def __repr__(self) -> str:
        if self.size is not None:
            meta = f"{self.size!r}"
        elif self.query_width is not None:
            meta = f"min-width: {self.query_width}px"
        else:
            meta = f"{type(self).__name__}"

        return f"{meta}; {self.type}"

    def __hash__(self) -> int:
        return hash(self.url)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ImageVersion):
            return NotImplemented
        return self.url == other.url

    def __lt__(self, other: "ImageVersion") -> bool:
        if isinstance(other, ImageVersion):
            if self.size and other.size and self.size != other.size:
                return self.size < other.size
            if (self.size is None) != (other.size is None):
                return self.size is None

            if self.query_width and other.query_width and self.query_width != other.query_width:
                return self.query_width < other.query_width
            if (self.query_width is None) != (other.query_width is None):
                return self.query_width is None

            if self.type and other.type and self.type != other.type:
                return self.type < other.type
            if (self.type is None) != (other.type is None):
                return self.type is None

            return self.url < other.url
        raise NotImplementedError(f"'<' is not defined between {type(self).__name__!r} and {type(other).__name__!r}")


@dataclass(frozen=False)
class Image(DataclassSerializationMixin):
    versions: List[ImageVersion]
    is_cover: bool
    description: Optional[str]
    caption: Optional[str]
    authors: List[str]
    position: int

    def __post_init__(self):
        for url in [version.url for version in self.versions]:
            if not validators.url(url, strict_query=False):
                raise ValueError(f"url {url} is not a valid URL")

    @property
    def url(self) -> str:
        return self.versions[-1].url

    def __str__(self) -> str:
        if self.is_cover:
            representation = "Fundus-Article Cover-Image:\n"
        else:
            representation = "Fundus-Article Image:\n"
        representation += (
            f"-URL:\t\t\t {self.url!r}\n"
            f"-Description:\t {self.description!r}\n"
            f"-Caption:\t\t {self.caption!r}\n"
            f"-Authors:\t\t {self.authors}\n"
            f"-Versions:\t\t {sorted(set(v.size for v in self.versions if v.size is not None))}\n"
        )
        return representation

    def __repr__(self) -> str:
        return self.url


class DOM:
    def __init__(self, root: lxml.html.HtmlElement):
        self.root = root
        self._depth_first_index = {element: i for i, element in enumerate(root.iter())}

    def get_index(self, node: lxml.html.HtmlElement) -> int:
        return self._depth_first_index[node]



import socket
import threading
from contextlib import contextmanager
from dataclasses import dataclass
from queue import Empty, Queue
from typing import Optional, Union

import requests.adapters

from fundus.logging import create_logger
from fundus.utils.events import __EVENTS__

logger = create_logger(__name__)

_default_header = {"user-agent": "Fundus/2.0 (contact: github.com/flairnlp/fundus)"}


class InterruptableSession(requests.Session):
    def __init__(self, timeout: Optional[int] = None):
        super().__init__()
        self.timeout = timeout

    def get_with_interrupt(self, *args, **kwargs) -> requests.Response:
        """Interruptable request.

        This function hands over the request to another thread and checks every second
        for an interrupt event. If there was an interrupt event, this function raises
        a requests.exceptions.Timeout error.

        Args:
            *args: requests.Session.get(*) arguments.
            **kwargs: requests.Session.get(**) keyword arguments.

        Returns:
            The response.
        """

        def _req():
            try:
                response_queue.put(self.get(*args, **kwargs, timeout=self.timeout))
            except Exception as error:
                response_queue.put(error)

        if args:
            url = args[0]
        else:
            url = kwargs.get("url")

        response_queue: Queue[Union[requests.Response, Exception]] = Queue()
        thread = threading.Thread(target=_req, daemon=True)
        thread.start()

        while True:
            try:
                response = response_queue.get(timeout=1)
            except Empty:
                if __EVENTS__.is_event_set("stop"):
                    logger.debug(f"Interrupt request for {url!r}")
                    response_queue.task_done()
                    exit(1)
            else:
                if isinstance(response, Exception):
                    raise response
                return response


@dataclass
class CONFIG:
    POOL_CONNECTIONS: int = 50
    POOL_MAXSIZE: int = 1
    TIMEOUT: Optional[int] = 30


class SessionHandler:
    """Object for handling project global request.Session

    The session life cycle consists of three steps which can be repeated indefinitely:
    Build, Supply, Teardown.
    Initially there is no session build within the session handler. When a session is requested
    with get_session() either a new one is created with _session_factory() or the session handler's
    existing one returned. Every subsequent call to get_session() will return the same
    response.Session object. If close_current_session() is called, the current session will be
    tear-downed and the next call to get_session() will build a new session.
    """

    CONFIG: CONFIG = CONFIG()

    def __init__(self):
        self.session: Optional[InterruptableSession] = None
        self.lock = threading.Lock()

    def _session_factory(self) -> InterruptableSession:
        """Builds a new Session

        This returns a new client session build from pre-defined configurations:
        - pool_connections: <self.pool_connections>
        - pool_maxsize: <self.pool_maxsize>
        - hooks: (1) Hook to raise an `HTTPError` if one occurred. (2) Hook to log the request responses.

        Returns:
            A new requests.Session
        """

        logger.debug("Creating new session")
        session = InterruptableSession(timeout=self.CONFIG.TIMEOUT)

        def _response_log(response: requests.Response, *args, **kwargs) -> None:
            history = response.history
            previous_status_codes = [f"({response.status_code})" for response in history] if history else []
            status_code_chain = " -> ".join(previous_status_codes + [f"({response.status_code})"])
            logger.debug(
                f"{status_code_chain} <{response.request.method} {response.url!r}> "
                f"took {response.elapsed.total_seconds()} second(s)"
            )

        # hooks
        response_hooks = [lambda response, *args, **kwargs: response.raise_for_status(), _response_log]
        session.hooks["response"].extend(response_hooks)

        # adapters
        session.mount(
            "http://",
            requests.adapters.HTTPAdapter(
                pool_connections=self.CONFIG.POOL_CONNECTIONS, pool_maxsize=self.CONFIG.POOL_MAXSIZE
            ),
        )
        session.mount(
            "https://",
            requests.adapters.HTTPAdapter(
                pool_connections=self.CONFIG.POOL_CONNECTIONS, pool_maxsize=self.CONFIG.POOL_MAXSIZE
            ),
        )

        return session

    def get_session(self) -> InterruptableSession:
        """Requests the current build session

        If called for the first time or after close_current_session was called,
        this function will build a new session. Every subsequent call will return
        the same session object until the session is closed with close_current_session().

        Returns:
            requests.Session: The current build session
        """

        with self.lock:
            if not self.session:
                self.session = self._session_factory()
            return self.session

    def close_current_session(self) -> None:
        """Tears down the current build session

        Returns:
            None
        """
        if self.session is not None:
            session = self.get_session()
            logger.debug(f"Close session {session}")
            session.close()
            self.session = None

    @classmethod
    @contextmanager
    def context(cls, **kwargs):
        """Context manager to temporarily overwrite session parameters.

        Returns:
            SessionHandler: The session handler instance.
        """

        cls.CONFIG = CONFIG(**kwargs)

        try:
            yield cls
        finally:
            cls.CONFIG = CONFIG()


@contextmanager
def socket_timeout(timeout: Optional[int] = None):
    """Temporarily sets the socket timeout within this context."""
    old_timeout = socket.getdefaulttimeout()
    socket.setdefaulttimeout(timeout)
    try:
        yield
    finally:
        socket.setdefaulttimeout(old_timeout)


session_handler = SessionHandler()



import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class TheGlobeAndMailParser(ParserProxy):
    class V1(BaseParser):
        _subheadline_selector = CSSSelector("article > h4")
        _paragraph_selector = CSSSelector("article > p")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def topics(self) -> List[str]:
            topic_list = [topic.lower() for topic in generic_topic_parsing(self.precomputed.meta.get("keywords"))]
            topic_set = set(topic_list)
            topic_duplicates = list(topic_list)
            for element in topic_set:
                topic_duplicates.remove(element)
            for duplicate in topic_duplicates:
                topic_list.remove(duplicate)
            return [topic.title() for topic in topic_list if "news" not in topic]

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                caption_selector=XPath("./ancestor::figure//figcaption//p[@class='figcap-text']/span[1]"),
                author_selector=XPath("./ancestor::figure//figcaption//p[@class='figcap-text']/span[last()]"),
            )



import datetime
from typing import Any, Dict

import pytest

from fundus import Article
from fundus.scraping.html import HTML, SourceInfo

info = SourceInfo(publisher="")
html = HTML(content="", responded_url="", requested_url="", crawl_date=datetime.datetime.now(), source_info=info)


class TestArticle:
    def test_constructor(self):
        extraction = {"authors": ["Author"], "title": "title"}

        with pytest.raises(TypeError):
            article = Article(extraction, html=html)  # type: ignore[arg-type, misc]

        with pytest.raises(TypeError):
            article = Article(**extraction)  # type: ignore[arg-type]

        article = Article(**{}, html=html)
        article = Article(**extraction, html=html, exception=None)
        article = Article(html=html, **extraction, exception=None)
        article = Article(**extraction, html=html, exception=TypeError())

    def test_default_values(self):
        extraction: Dict[str, Any] = {}

        article = Article(**extraction, html=html)

        assert article.title is None
        assert article.body is None
        assert article.authors == []
        assert article.publishing_date is None
        assert article.topics == []
        assert article.free_access is False

    def test_view(self):
        extraction = {
            "authors": ["Author1", "Author2", "Author3"],
            "title": "<TITLE>",
        }

        article = Article(**extraction, html=html, exception=None)

        assert article.title == "<TITLE>"
        assert article.authors == ["Author1", "Author2", "Author3"]

    def test_extraction_view_getter(self):
        extraction = {"test_attribute": "test_value"}

        article = Article(**extraction, html=html, exception=None)

        assert article.test_attribute
        assert article.test_attribute == "test_value"

        article.__extraction__["test_attribute"] = "very_secret_stuff"  # type: ignore[index]

        assert article.test_attribute == "very_secret_stuff"

    def test_extraction_view_setter(self):
        extraction = {"test_attribute": "test_value"}

        article = Article(**extraction, html=html, exception=None)
        with pytest.raises(AttributeError):
            article.test_attribute = "another_value"



from datetime import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class TheNewYorkerParser(ParserProxy):
    class V1(BaseParser):
        _summary_selector = XPath("//div[contains(@class, 'ContentHeaderDek')]")
        _paragraph_selector = CSSSelector("div.body__inner-container > p")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute(validate=False)
        def description(self) -> Optional[str]:
            return self.precomputed.meta.get("og:description")

        @attribute(validate=False)
        def alternative_description(self) -> Optional[str]:
            return self.precomputed.ld.xpath_search("NewsArticle/description", scalar=True)

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.xpath_search("NewsArticle/author"))

        @attribute
        def publishing_date(self) -> Optional[datetime]:
            return generic_date_parsing(self.precomputed.ld.xpath_search("NewsArticle/datePublished", scalar=True))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.xpath_search("NewsArticle/headline", scalar=True)

        @attribute(validate=False)
        def alternative_title(self) -> Optional[str]:
            return self.precomputed.ld.xpath_search("NewsArticle/alternativeHeadline", scalar=True)

        @attribute
        def topics(self) -> List[str]:
            # The New Yorker has keywords in the meta as well as the ld+json.
            # Since the keywords from the meta seem of higher quality, we use these.
            # Example:
            # meta:    ['the arctic', 'ice', 'climate change']
            # ld+json: ['the control of nature', 'the arctic', 'ice', 'climate change', 'splitscreenimagerightinset', 'web']
            return generic_topic_parsing(self.precomputed.meta.get("keywords"))

        @attribute(validate=False)
        def section(self) -> Optional[str]:
            return self.precomputed.ld.xpath_search("NewsArticle/articleSection", scalar=True)

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//picture//img"),
                caption_selector=XPath(
                    "./ancestor::*[self::figure or self::header]//*[(self::span and contains(@class, 'caption__text')) or (self::div and contains(@class, '__caption'))]"
                ),
                author_selector=XPath(
                    "(./ancestor::*[self::figure or self::header]//*[(self::span and contains(@class, 'caption__credit')) or (self::div and contains(@class, '__credit'))])[last()]"
                ),
            )



import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class TheMirrorParser(ParserProxy):
    class V1(BaseParser):
        VALID_UNTIL = datetime.date(2024, 7, 26)
        _paragraph_selector = XPath(
            "/html/body/main/article/div[@class='article-body']/p[text()] | //div[@class='article-body']//div[@class='live-event-lead-entry']/p[text()] | //div[@class='article-body']//div[@class='entry-content']/p[text()]"
        )
        _summary_selector = XPath("/html/body/main/article/div[@class='lead-content']/p")
        _subheadline_selector = XPath(
            "//div[@class='article-body']/h3 | //div[@class='article-body']//div[@class='entry-content']/h3"
        )

        @attribute
        def body(self) -> Optional[ArticleBody]:
            body = extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                paragraph_selector=self._paragraph_selector,
                subheadline_selector=self._subheadline_selector,
            )
            return body

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.meta.get("parsely-pub-date"))

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.meta.get("author"))

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("keywords"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=CSSSelector("div.image > img, div.image-container amp-img"),
                caption_selector=XPath(
                    "./ancestor::div[@class='lead-content' or @class='image-container']//figcaption//span[1]"
                ),
                author_selector=XPath(
                    "./ancestor::div[@class='lead-content' or @class='image-container']//figcaption//span[2]"
                ),
                lower_boundary_selector=CSSSelector("reach-viafoura-comments"),
            )

    class V1_1(V1):
        VALID_UNTIL = datetime.date.today()

        _datetime_selector = CSSSelector("div.article-information[itemprop='datePublished']")

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            if date_nodes := self._datetime_selector(self.precomputed.doc):
                return generic_date_parsing(date_nodes[0].attrib.get("content"))
            return None



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.publishers.cz.seznamzpravy import SeznamZpravyParser
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap


class CZ(metaclass=PublisherGroup):
    default_language = "cs"

    SeznamZpravy = Publisher(
        name="SeznamZpravy",
        domain="https://seznamzpravy.cz/",
        parser=SeznamZpravyParser,
        sources=[
            RSSFeed("https://www.seznamzpravy.cz/rss"),
            Sitemap("https://www.seznamzpravy.cz/sitemaps/sitemap_articles.xml"),
            NewsMap("https://www.seznamzpravy.cz/sitemaps/sitemap_news.xml"),
        ],
    )



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, ParserProxy, attribute
from fundus.parser.data import Image
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)
from fundus.scraping.filter import regex_filter


class NationalPostParser(ParserProxy):
    class V1(BaseParser):
        # Note: this date is approximate. The actual date lies between 2025-04-06 and 2025-04-15. This is due to
        # National Posts exclusion from archive.org and problems with the publisher coverage.
        VALID_UNTIL = datetime.date(2025, 4, 15)

        _summary_selector = CSSSelector("article p.article-subtitle")
        _subheadline_selector = XPath(
            "//section[@class='article-content__content-group article-content__content-group--story']/p/strong | "
            "//section[@class='article-content__content-group article-content__content-group--story']/h3"
        )
        _paragraph_selector = XPath(
            "//section[@class='article-content__content-group article-content__content-group--story']/p[text()]"
        )

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def topics(self) -> List[str]:
            preliminary_topics = self.precomputed.ld.bf_search("keywords")
            filter_list = ["Curated", "News", "Newsroom daily", "story", "Canada", "World", "nationalpost.com"]
            topic_filter = regex_filter(
                r"([0-9a-f]{8}-([0-9a-f]{4}-){3}[0-9a-f]{12}|NLP Entity Tokens|NLP Category|NP Comment|Category:)"
            )
            filtered_topics = [
                topic for topic in preliminary_topics if not topic_filter(topic) and topic not in filter_list
            ]
            return generic_topic_parsing(filtered_topics)

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath("//div[@class='article-header__detail']/figure"),
                lower_boundary_selector=CSSSelector("section.article-delimiter"),
            )

    class V1_1(V1):
        VALID_UNTIL = datetime.date.today()

        _paragraph_selector = XPath(
            "//div[@class='story-v2-content-element-inline']/"
            "p[text() and not(@data-async) and not(text()='National Post')]"
        )
        _subheadline_selector = XPath(
            "//div[@class='story-v2-content-element-inline']/h3 |"
            "//div[@class='story-v2-content-element-inline']/p/strong"
        )

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath("(//div[@class='story-v2-block story-v2-article-container'])[1]"),
                lower_boundary_selector=XPath("//section[@class='article-content__share-group']"),
                caption_selector=XPath("./ancestor::figure/figcaption/span[@class='caption']"),
                author_selector=XPath("./ancestor::figure/figcaption/span[@class='credit' or @class='distributor']"),
            )



import re
from datetime import date, datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import (
    ArticleBody,
    BaseParser,
    Image,
    ParserProxy,
    attribute,
    utility,
)
from fundus.parser.utility import image_extraction


class DailyNewsTZParser(ParserProxy):
    class V1(BaseParser):
        # VALID_UNTIL = date(2025, 5, 3)
        _summary_selector = CSSSelector("div.cs-entry__subtitle")
        _subheadline_selector = XPath(
            "//div[contains(@class,'entry-content')]//p[not(text() or position()=1)]//span//strong"
        )
        _paragraph_selector = XPath(
            "//div[contains(@class, 'entry-content')]"
            "//p[not(re:test(string(.), '^(SOMA|ALSO READ):') or span or @class) and text()] | "
            "//div[contains(@class, 'entry-content')]//p[not(position()=1 or @class)]//span[not(span) and text()] |"
            "//div[contains(@class, 'entry-content')]//p[not(@class)]//span/span[text()] | "
            "//div[contains(@class, 'entry-content')]//p[position()=1 and not(@class or a)] | "
            "//div[contains(@class, 'entry-content')]//span[@data-offset-key]",
            namespaces={"re": "http://exslt.org/regular-expressions"},
        )

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.xpath_search("//Article//headline", scalar=True) or re.sub(
                r"(?i)\s*-\s*(daily\s*news|habari\s*leo)\s*", "", self.precomputed.meta.get("og:title") or ""
            )

        @attribute
        def body(self) -> Optional[ArticleBody]:
            article_body = utility.extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )
            return article_body

        @attribute
        def authors(self) -> List[str]:
            return utility.generic_author_parsing(
                self.precomputed.meta.get("twitter:data1") or self.precomputed.ld.xpath_search("//Article//author")
            )

        @attribute
        def publishing_date(self) -> Optional[datetime]:
            return utility.generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath("//div[@id='content']"),
                image_selector=XPath("//figure//img[1]|//div[@id='content']//p/img"),
                caption_selector=XPath(
                    "./ancestor::figure//figcaption | "
                    "./ancestor::div[@class='cs-entry__thumbnail']//div[@class='cs-entry__thumbnail-caption'] |"
                    "(./ancestor::p//following-sibling::p[@style='text-align: center'])[1]/strong"
                ),
                author_selector=re.compile(r"\((?P<credits>[^()]+)\)"),
            )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap

from .nine_news import NineNewsParser
from .west_australian import WestAustralianParser


class AU(metaclass=PublisherGroup):
    default_language = "en"

    NineNews = Publisher(
        name="Nine News",
        domain="https://www.9news.com.au/",
        parser=NineNewsParser,
        sources=[
            RSSFeed("https://www.9news.com.au/rss"),
            Sitemap(
                "https://www.9news.com.au/sitemap.xml",
                sitemap_filter=inverse(regex_filter("sitemap-content-")),
            ),
        ],
    )

    WestAustralian = Publisher(
        name="The West Australian",
        domain="https://thewest.com.au/",
        parser=WestAustralianParser,
        sources=[
            RSSFeed("https://thewest.com.au/rss"),
            NewsMap("https://thewest.com.au/news-sitemap.xml"),
            Sitemap("https://thewest.com.au/sitemap.xml", reverse=True),
        ],
    )



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class INewsParser(ParserProxy):
    class V1(BaseParser):
        VALID_UNTIL = datetime.date(2025, 1, 1)
        _summary_selector = CSSSelector("article > h2")
        _paragraph_selector = CSSSelector("article div.article-content p")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            body = extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                paragraph_selector=self._paragraph_selector,
            )
            return body

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.ld.bf_search("keywords"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=CSSSelector("div.inews__main"),
                image_selector=CSSSelector("figure:has(> figcaption) img"),
                author_selector=re.compile(r"\((?P<credits>.*?)\)$"),
            )

    class V1_1(V1):
        VALID_UNTIL = datetime.date.today()

        _summary_selector = CSSSelector("article p.inews__post-excerpt")

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=CSSSelector("div.article-wrapper"),
                image_selector=CSSSelector("figure:has(> figcaption) img"),
                author_selector=re.compile(r"\((?P<credits>.*?)\)$"),
            )



import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, ParserProxy, attribute
from fundus.parser.data import Image
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class BerlinerZeitungParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = CSSSelector("div[id=articleBody] > p")
        _summary_selector = CSSSelector("div[data-testid=article-header] > p")
        _subheadline_selector = CSSSelector("div[id=articleBody] > h2")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                subheadline_selector=self._subheadline_selector,
                summary_selector=self._summary_selector,
            )

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.meta.get("article:author"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.ld.bf_search("keywords"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//div[not(contains(@class, 'author') or contains(@class, 'preview'))]/img"),
                caption_selector=XPath(
                    "./ancestor::div[@class='article_image-container__Yo6Cx']"
                    "//span[@class='article_image-container-caption__lZ5kc']"
                ),
                author_selector=XPath(
                    "./ancestor::div[@class='article_image-container__Yo6Cx']"
                    "//span[@class='article_image-container-source__rbsO4']"
                ),
            )



import re
from datetime import date, datetime
from typing import List, Optional

import lxml.html
from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import (
    ArticleBody,
    BaseParser,
    Image,
    ParserProxy,
    attribute,
    function,
)
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class TheNationParser(ParserProxy):
    class V1(BaseParser):
        VALID_UNTIL = date(2023, 7, 22)

        # There is a known issue preventing lxml from extracting text content within
        # the specified summary node.
        # This is due to invalid XML provided by The Nation.
        # Currently(lxml 4.9.3), lxml does not accept p tags within any heading (h*) tag.
        # The "correct" selector would be ".article-header-content > h2 > p"
        _summary_selector: XPath = CSSSelector(".article-header-content > h2")
        _paragraph_selector: XPath = CSSSelector(".article-body-inner > p")
        _aside_selector = CSSSelector("aside")

        _html_fix_pattern = re.compile(r'name="sft_double_opt_sail"\s*value="yes"\s/>\s*</form>')

        # there is a known issue with broken HTML regarding this publisher.
        # div.cta subtree is malformed, there is a missing closing div tag before the </form> tag, which
        # prevents lxml from properly parsing the text. I don't know if this is also a problem within V1,
        # but better safe than sorry, so I added it to V1 base class
        @function(priority=1)
        def _fix_malformed_html(self) -> None:
            if self.precomputed.doc.xpath("//div[contains(@id, 'cta-block')]"):
                fixed_html = re.sub(
                    self._html_fix_pattern,
                    'name="sft_double_opt_sail"value="yes"/></div></form>',
                    self.precomputed.html,
                )
                self.precomputed.doc = lxml.html.document_fromstring(fixed_html)

        # We remove aside tags here because the provided HTML does not enclose <p> tags
        # within .article-header-content. As a result, <aside> tags following <p> tags get attached
        # to the paragraph. This is valid HTML5 behaviour.
        # see https://stackoverflow.com/questions/8460993/p-end-tag-p-is-not-needed-in-html
        @function(priority=3)
        def _remove_aside(self) -> None:
            for aside in self._aside_selector(self.precomputed.doc):
                if (parent := aside.getparent()) is not None:
                    parent.remove(aside)

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.meta.get("sailthru.author"))

        @attribute
        def publishing_date(self) -> Optional[datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("keywords"))

        @attribute(priority=2)
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath("//h1[contains(@class,'title')]"),
                image_selector=CSSSelector(".image img"),
                caption_selector=XPath("(./ancestor::aside[contains(@class, 'image')])[1]//p[@class='caption']/text()"),
                author_selector=XPath(
                    "(./ancestor::aside[contains(@class, 'image')])[1]//p[@class='caption']/span[@class='credits']"
                ),
            )

    class V2(V1):
        VALID_UNTIL = date.today()

        # oh boy, TheNation is really a mess. they changed the layout 2023|7|22 but somehow the old articles still
        # use the old layout for main content, so we concatenate XPath from V1 onto V1_1.

        _summary_selector = XPath(
            "//div[@class='article-header-content'] /h2 | //div[contains(@class, 'article-title')] /p"
        )
        _paragraph_selector = XPath("(//article | //div[@class='article-body-inner']) / p")

        # remove aside function from V1
        def _remove_aside(self):
            pass

        @attribute
        def topics(self) -> List[str]:
            if topics := generic_topic_parsing(self.precomputed.meta.get("keywords")):
                return topics
            else:
                return generic_topic_parsing(self.precomputed.meta.get("sailthru.tags"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath("//h1[contains(@class,'title')]"),
                caption_selector=XPath("./ancestor::figure//figcaption/text()|./ancestor::figure//figcaption/p"),
            )



import datetime
import re
from typing import List, Optional, Pattern

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    apply_substitution_pattern_over_list,
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class DieZeitParser(ParserProxy):
    class V1(BaseParser):
        _author_substitution_pattern: Pattern[str] = re.compile(r"DIE ZEIT (Archiv)")
        _paragraph_selector = XPath("//div[@class = 'article-page']/p[not(contains(text(), '© dpa-infocom'))]")
        _summary_selector = CSSSelector("div.summary")
        _subheadline_selector = CSSSelector("div.article-page > h2")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return apply_substitution_pattern_over_list(
                generic_author_parsing(self.precomputed.ld.bf_search("author")), self._author_substitution_pattern
            )

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("keywords"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//figure//img[@class='article__media-item']"),
                caption_selector=XPath("./ancestor::figure//span[@class='figure__text']"),
                author_selector=XPath("./ancestor::figure//span[@class='figure__copyright']"),
                lower_boundary_selector=XPath("//nav[@class='breadcrumbs']"),
            )
